@@ -30,10 +30,8 @@ async fn main() -> Result<(), Box<dyn Error>> {
 
 		println!();
 		for other_utilization_descriptor in all_utilizations.object_descriptors() {
-			if let Some(state) = all_utilizations.object_state(&other_utilization_descriptor)? {
-				if let Some(utilization) = state.data {
-					println!("{} -> {:?}", other_utilization_descriptor.json.get("id").unwrap().as_str().unwrap(), utilization.percentage);
-				}
+			if let Some(utilization) = all_utilizations.object_state(&other_utilization_descriptor)? {
+				println!("{} -> {:?}", other_utilization_descriptor.json.get("id").unwrap().as_str().unwrap(), utilization.data.percentage);
 			}
 		}
 
