@@ -21,6 +21,20 @@ impl ObjectDescriptor {
 		let json: serde_json::value::Value = json.into();
 		ObjectDescriptor { tags: tags.into_iter().map(|descriptor| TagDescriptor { json: descriptor.into() }).collect(), json }
 	}
+
+	pub fn json_to_string(&self) -> String {
+		self.json.to_string()
+	}
+
+	pub fn tags_to_strings(&self) -> Vec<String> {
+		self.tags.iter().map(|tag| tag.json_to_string()).collect()
+	}
+
+	pub fn from_json_strings(tags: Vec<String>, json: String) -> ObjectDescriptor {
+		let tags: HashSet<TagDescriptor> = tags.into_iter().map(TagDescriptor::from_json_string).collect();
+		let json: serde_json::value::Value = serde_json::from_str(json.as_str()).unwrap();
+		ObjectDescriptor { tags, json }
+	}
 }
 
 impl TagDescriptor {
@@ -28,6 +42,15 @@ impl TagDescriptor {
 		let json: serde_json::value::Value = value.into();
 		TagDescriptor { json }
 	}
+
+	pub fn json_to_string(&self) -> String {
+		self.json.to_string()
+	}
+
+	pub fn from_json_string(json: String) -> TagDescriptor {
+		let json: serde_json::value::Value = serde_json::from_str(json.as_str()).unwrap();
+		TagDescriptor { json }
+	}
 }
 
 
@@ -35,7 +58,7 @@ impl TagDescriptor {
 use itertools::Itertools;
 
 
-fn hash_serde_json_value<H: Hasher>(value: &serde_json::Value, state: &mut H) {
+pub(crate) fn hash_serde_json_value<H: Hasher>(value: &serde_json::Value, state: &mut H) {
 	match value {
 		serde_json::Value::Null => {
 			state.write_u8(0);
@@ -126,25 +149,25 @@ impl Hash for TagDescriptor {
 
 impl std::fmt::Debug for ObjectDescriptor {
 	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-		f.write_fmt(format_args!("ObjectDescriptor:Tags{:?}:JSON{}", self.tags, self.json.to_string()))
+		f.write_fmt(format_args!("ObjectDescriptor:Tags{:?}:JSON{}", self.tags, self.json))
 	}
 }
 
 impl std::fmt::Display for ObjectDescriptor {
 	fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-		f.write_fmt(format_args!("{}", self.json.to_string()))
+		f.write_fmt(format_args!("{}", self.json))
 	}
 }
 
 impl std::fmt::Debug for TagDescriptor {
 	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-		f.write_fmt(format_args!("TagDescriptor:{}", self.json.to_string()))
+		f.write_fmt(format_args!("TagDescriptor:{}", self.json))
 	}
 }
 
 impl std::fmt::Display for TagDescriptor {
 	fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-		f.write_fmt(format_args!("{}", self.json.to_string()))
+		f.write_fmt(format_args!("{}", self.json))
 	}
 }
 
