@@ -463,7 +463,7 @@ fn symbol_at(input: &[u8], location: Position) -> u32 {
 }
 
 fn symbol_to_str(symbol: u32) -> String {
-	let bytes: [u8; POSITION_BYTES] = unsafe { std::intrinsics::transmute(symbol) }; //TODO function
+	let bytes: [u8; POSITION_BYTES] = symbol.to_ne_bytes();
 	String::from_utf8_lossy(&bytes).into_owned()
 }
 
