@@ -13,7 +13,7 @@ TO-DOCUMENT:
 
 // TODO: split this into multiple files
 
-use std::{collections::HashMap, fmt::Debug, hash::Hash, sync::{Arc, Condvar, Mutex, RwLock}, task::{Context, Poll, Waker}};
+use std::{collections::HashMap, fmt::Debug, hash::Hash, sync::{Arc, Condvar, Mutex, RwLock}, task::Waker};
 
 
 pub trait CallbackRegistry: Send + Sync {
@@ -662,14 +662,14 @@ impl<K: Key> Iterator for EventIterator<K> {
 impl<K: Key> futures::stream::Stream for EventStream<K> {
 	type Item = Event<K>;
 
-	fn poll_next(self: std::pin::Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
+	fn poll_next(self: std::pin::Pin<&mut Self>, cx: &mut std::task::Context<'_>) -> std::task::Poll<Option<Self::Item>> {
 		let mut notifications = self.shared.events.lock().unwrap();
 		if let Some(key) = notifications.keys().next().cloned() {
 			*self.shared.waker.write().unwrap() = None;
-			return Poll::Ready(Some(Event { action: notifications.remove(&key).unwrap(), key }));
+			return std::task::Poll::Ready(Some(Event { action: notifications.remove(&key).unwrap(), key }));
 		};
 		*self.shared.waker.write().unwrap() = Some(cx.waker().clone());
-		Poll::Pending
+		std::task::Poll::Pending
 	}
 }
 
