@@ -1,10 +1,10 @@
 #![allow(clippy::not_unsafe_ptr_arg_deref)]
-use std::{convert::TryFrom, ffi::c_void, mem::forget, sync::Arc};
+use std::{convert::TryFrom, mem::forget, sync::Arc};
 
 use log::error;
 
 use super::FFIObjectDescriptor;
-use crate::{events::{Callback, CallbackRegistry, Event}, ffi::{make_object_descriptor_callback, FFIError, ObjectEventCallbackFn}, object::core::{ObjectRawData, ObjectType, ObjectVersion}, LocalNode, ObjectDescriptor, ObjectExpose};
+use crate::{events::{CallbackRegistry, Event}, ffi::FFIError, object::core::{ObjectRawData, ObjectType, ObjectVersion}, LocalNode, ObjectDescriptor, ObjectExpose};
 
 
 pub(super) fn raw_serialize(cooked: &(ObjectType, Vec<u8>)) -> Result<(ObjectType, ObjectRawData), ()> {
@@ -74,9 +74,10 @@ pub struct FFIObjectExposeStateResult {
 	pub changed: u8,
 }
 
+//TODO: expose AT
 #[no_mangle]
 pub extern "C" fn hakuban_object_expose_state(
-	object_expose_ptr: *mut ObjectExpose<(Vec<String>, Vec<u8>), ()>, state: FFIObjectExposeState,
+	object_expose_ptr: *mut ObjectExpose<(Vec<String>, Vec<u8>), ()>, state: FFIObjectExposeState, assignment_id: u64,
 ) -> FFIObjectExposeStateResult {
 	let (data, data_type, data_version) = match state.to_rust() {
 		Ok(state) => state,
@@ -84,7 +85,7 @@ pub extern "C" fn hakuban_object_expose_state(
 	};
 	let cooked = (data_type, data);
 	let object_expose = unsafe { Box::from_raw(object_expose_ptr) };
-	let changed = object_expose.set_object_state(&data_version, &cooked).unwrap();
+	let changed = object_expose.set_object_state(&data_version, &cooked, assignment_id).unwrap();
 	forget(object_expose);
 	forget(cooked.1);
 	forget(data_version);
@@ -93,19 +94,27 @@ pub extern "C" fn hakuban_object_expose_state(
 
 
 #[no_mangle]
-pub extern "C" fn hakuban_object_expose_assigned(object_expose_ptr: *mut ObjectExpose<(Vec<String>, Vec<u8>), ()>) -> u8 {
+pub extern "C" fn hakuban_object_expose_assignment(object_expose_ptr: *mut ObjectExpose<(Vec<String>, Vec<u8>), ()>) -> u64 {
 	let object_expose = unsafe { Box::from_raw(object_expose_ptr) };
-	let ret = if object_expose.assigned() { 1 } else { 0 };
+	let ret = if let Some(assignment) = object_expose.assignment() { assignment } else { 0 };
 	forget(object_expose);
 	ret
 }
 
 #[no_mangle]
-pub extern "C" fn hakuban_object_expose_changes_callback_register(
-	object_expose_ptr: *mut ObjectExpose<ObjectRawData, ()>, callback: ObjectEventCallbackFn, userdata: *const c_void,
-) -> *mut Box<dyn Callback<Event<ObjectDescriptor>>> {
+pub extern "C" fn hakuban_object_expose_desynchronize(object_expose_ptr: *mut ObjectExpose<(Vec<String>, Vec<u8>), ()>, assignment_id: u64) {
 	let object_expose = unsafe { Box::from_raw(object_expose_ptr) };
-	let ret = Box::new(object_expose.changes().register(make_object_descriptor_callback(callback, userdata)));
+	object_expose.desynchronize(assignment_id);
+	forget(object_expose);
+}
+
+
+#[no_mangle]
+pub extern "C" fn hakuban_object_expose_events_get(
+	object_expose_ptr: *mut ObjectExpose<ObjectRawData, ()>,
+) -> *mut Box<dyn CallbackRegistry<Params = Event<ObjectDescriptor>>> {
+	let object_expose = unsafe { Box::from_raw(object_expose_ptr) };
+	let ret = Box::new(Box::new(object_expose.changes()) as Box<dyn CallbackRegistry<Params = Event<ObjectDescriptor>>>);
 	forget(object_expose);
 	Box::into_raw(ret)
 }
