@@ -1,8 +1,8 @@
 #![allow(clippy::not_unsafe_ptr_arg_deref)]
-use std::{convert::TryFrom, ffi::c_void, mem::forget};
+use std::{convert::TryFrom, mem::forget};
 
 use super::{raw_serialize, FFIObjectDescriptor, FFIObjectDescriptors, FFIObjectExposeState, FFIObjectExposeStateResult, FFITagDescriptor};
-use crate::{events::{Callback, CallbackRegistry, Event}, ffi::{make_object_descriptor_callback, FFIError, ObjectEventCallbackFn}, object::core::ObjectRawData, LocalNode, ObjectDescriptor, TagDescriptor, TagExpose};
+use crate::{events::{CallbackRegistry, Event}, ffi::FFIError, object::core::ObjectRawData, LocalNode, ObjectDescriptor, TagDescriptor, TagExpose};
 
 #[repr(C)]
 pub struct FFITagExposeResult {
@@ -30,9 +30,10 @@ pub extern "C" fn hakuban_tag_expose_drop(tag_expose_ptr: *mut TagExpose<ObjectR
 
 
 
+//TODO: expose AT
 #[no_mangle]
 pub extern "C" fn hakuban_tag_expose_object_state(
-	tag_expose_ptr: *mut TagExpose<(Vec<String>, Vec<u8>), ()>, descriptor: FFIObjectDescriptor, state: FFIObjectExposeState,
+	tag_expose_ptr: *mut TagExpose<(Vec<String>, Vec<u8>), ()>, descriptor: FFIObjectDescriptor, state: FFIObjectExposeState, assignment_id: u64,
 ) -> FFIObjectExposeStateResult {
 	let descriptor: ObjectDescriptor = match ObjectDescriptor::try_from(&descriptor) {
 		Ok(descriptor) => descriptor,
@@ -44,7 +45,7 @@ pub extern "C" fn hakuban_tag_expose_object_state(
 	};
 	let cooked = (data_type, data);
 	let tag_expose = unsafe { Box::from_raw(tag_expose_ptr) };
-	let changed = tag_expose.set_object_state(&descriptor, &data_version, &cooked).unwrap();
+	let changed = tag_expose.set_object_state(&descriptor, &data_version, &cooked, assignment_id).unwrap();
 	forget(tag_expose);
 	forget(cooked.1);
 	forget(data_version);
@@ -60,11 +61,36 @@ pub extern "C" fn hakuban_tag_expose_object_descriptors_borrow(tag_expose_ptr: *
 }
 
 #[no_mangle]
-pub extern "C" fn hakuban_tag_expose_objects_changes_callback_register(
-	tag_expose_ptr: *mut TagExpose<ObjectRawData, ()>, callback: ObjectEventCallbackFn, userdata: *const c_void,
-) -> *mut Box<dyn Callback<Event<ObjectDescriptor>>> {
+pub extern "C" fn hakuban_tag_expose_object_assignment(tag_expose_ptr: *mut TagExpose<(Vec<String>, Vec<u8>), ()>, descriptor: FFIObjectDescriptor) -> u64 {
+	let descriptor: ObjectDescriptor = match ObjectDescriptor::try_from(&descriptor) {
+		Ok(descriptor) => descriptor,
+		Err(error) => panic!("Invalid desciptor: {:?}", error),
+	};
 	let tag_expose = unsafe { Box::from_raw(tag_expose_ptr) };
-	let ret = Box::new(tag_expose.changes().register(make_object_descriptor_callback(callback, userdata)));
+	let ret = if let Some(assignment) = tag_expose.object_assignment(&descriptor) { assignment } else { 0 };
+	forget(tag_expose);
+	ret
+}
+
+#[no_mangle]
+pub extern "C" fn hakuban_tag_expose_object_desynchronize(
+	tag_expose_ptr: *mut TagExpose<(Vec<String>, Vec<u8>), ()>, descriptor: FFIObjectDescriptor, assignment_id: u64,
+) {
+	let descriptor: ObjectDescriptor = match ObjectDescriptor::try_from(&descriptor) {
+		Ok(descriptor) => descriptor,
+		Err(error) => panic!("Invalid desciptor: {:?}", error),
+	};
+	let tag_expose = unsafe { Box::from_raw(tag_expose_ptr) };
+	tag_expose.object_desynchronize(&descriptor, assignment_id);
+	forget(tag_expose);
+}
+
+#[no_mangle]
+pub extern "C" fn hakuban_tag_expose_events_get(
+	tag_expose_ptr: *mut TagExpose<ObjectRawData, ()>,
+) -> *mut Box<dyn CallbackRegistry<Params = Event<ObjectDescriptor>>> {
+	let tag_expose = unsafe { Box::from_raw(tag_expose_ptr) };
+	let ret = Box::new(Box::new(tag_expose.changes()) as Box<dyn CallbackRegistry<Params = Event<ObjectDescriptor>>>);
 	forget(tag_expose);
 	Box::into_raw(ret)
 }
