@@ -1,8 +1,8 @@
 #![allow(clippy::not_unsafe_ptr_arg_deref)]
-use std::{convert::TryFrom, ffi::c_void, mem::forget};
+use std::{convert::TryFrom, mem::forget, sync::Mutex};
 
-use super::{FFIObjectDescriptor, FFITagDescriptor};
-use crate::{events::{Callback, CallbackRegistry, Event}, ffi::{make_object_descriptor_callback, raw_deserialize, FFIError, FFIObjectObserveState, ObjectEventCallbackFn}, object::core::ObjectRawData, LocalNode, ObjectDescriptor, TagDescriptor, TagObserve};
+use super::{FFIObjectDescriptor, FFIObjectObserveState, FFITagDescriptor};
+use crate::{events::{CallbackRegistry, Event}, ffi::{raw_deserialize, FFIError}, object::core::ObjectRawData, LocalNode, ObjectDescriptor, TagDescriptor, TagObserve};
 
 #[repr(C)]
 pub struct FFITagObserveResult {
@@ -31,7 +31,7 @@ pub extern "C" fn hakuban_tag_observe_drop(tag_ptr: *mut TagObserve<ObjectRawDat
 #[repr(C)]
 pub struct FFITagObserveObjectStateBorrowResult {
 	pub error: FFIError,
-	pub state: FFIObjectObserveState,
+	pub state: *const Mutex<FFIObjectObserveState>,
 }
 
 #[no_mangle]
@@ -40,17 +40,21 @@ pub extern "C" fn hakuban_tag_observe_object_state_borrow(
 ) -> FFITagObserveObjectStateBorrowResult {
 	let descriptor: ObjectDescriptor = match ObjectDescriptor::try_from(&descriptor) {
 		Ok(descriptor) => descriptor,
-		Err(error) => return FFITagObserveObjectStateBorrowResult { error, state: FFIObjectObserveState::empty() },
+		Err(error) => return FFITagObserveObjectStateBorrowResult { error, state: std::ptr::null() },
 	};
 	let tag_observe = unsafe { Box::from_raw(tag_observe_ptr) };
 	let ret = match tag_observe.object_state(&descriptor).unwrap() {
-		Some(state) => FFITagObserveObjectStateBorrowResult { error: FFIError::None, state: FFIObjectObserveState::from_state(state) },
-		None => FFITagObserveObjectStateBorrowResult { error: FFIError::ObjectNotFound, state: FFIObjectObserveState::empty() },
+		Some(state) => FFITagObserveObjectStateBorrowResult {
+			error: FFIError::None,
+			state: Box::into_raw(Box::new(Mutex::new(FFIObjectObserveState { state, data_type_c_strings: None }))),
+		},
+		None => FFITagObserveObjectStateBorrowResult { error: FFIError::ObjectNotFound, state: std::ptr::null() },
 	};
 	forget(tag_observe);
 	ret
 }
 
+
 #[repr(C)]
 pub struct FFIObjectDescriptors {
 	pub count: usize,
@@ -86,12 +90,13 @@ pub extern "C" fn hakuban_object_descriptors_return(descriptors: FFIObjectDescri
 	descriptors.free();
 }
 
+
 #[no_mangle]
-pub extern "C" fn hakuban_tag_observe_objects_changes_callback_register(
-	tag_observe_ptr: *mut TagObserve<ObjectRawData, ()>, callback: ObjectEventCallbackFn, userdata: *const c_void,
-) -> *mut Box<dyn Callback<Event<ObjectDescriptor>>> {
+pub extern "C" fn hakuban_tag_observe_events_get(
+	tag_observe_ptr: *mut TagObserve<ObjectRawData, ()>,
+) -> *mut Box<dyn CallbackRegistry<Params = Event<ObjectDescriptor>>> {
 	let tag_observe = unsafe { Box::from_raw(tag_observe_ptr) };
-	let ret = Box::new(tag_observe.changes().register(make_object_descriptor_callback(callback, userdata)));
+	let ret = Box::new(Box::new(tag_observe.changes()) as Box<dyn CallbackRegistry<Params = Event<ObjectDescriptor>>>);
 	forget(tag_observe);
 	Box::into_raw(ret)
 }
