@@ -102,7 +102,7 @@ let _upstream = WebsocketConnector::new(hakuban.clone(), "ws://127.0.0.1:3001")?
 let expose = hakuban.object((["some-tag"],"xxx")).expose();
 
 for _event in expose.changes() {
-	if expose.assigned() {
+	if expose.assignment().is_some() {
 		expose.set_object_data(&"aaaaaa");
 		break;
 	}
@@ -183,24 +183,29 @@ It accepts connections and will naturally route hakuban communication, like ever
 */
 
 /*
-# TO-DO
 
 * features
 	* everything listed in the README/roadmap
+	* re-use diffs obtained from exposers
+	* gap-less expose handover
+	* avoid resending descriptors
 	* Events should have Emitter and Receiver, and Emitter should auto-unsubscribe on destroy. + Blanket implementation of Callback Registry.
 	* error!() in ffi functions, before better error handling gets implemented
 
 * clean-up
 	* check if error returns in ffi don't cut through allocations
 	* de arc-ify selfs where possible
-	* split events.rb into multiple files
+	* split events.rs into multiple files
 	* merge *contract.rs into a single file
-	* custom test harness https://www.fluvio.io/blog/2021/04/rust-custom-test-harness/
-	* return proper errors instead of Result<_, String>
+	* custom test harness; https://www.fluvio.io/blog/2021/04/rust-custom-test-harness/
+	* return proper errors instead of Result<_, String>; https://docs.rs/thiserror/latest/thiserror/index.html
 	* ensure all structs hashable by pointers are Pin
 	* remove superfluous fields from test yamls
 	* C-NEWTYPE: use struct ObjectVersion(Vec<i64>);  instead of type =, etc.
+	* remove unused functions, and struct fields
 	* C-DEBUG: implement debug on all public types
+	* change js descriptor hashing to secure
+	* unify state_set/set_state order
 
 * consider
 	* rename ObjectObserve -> ObjectObserveContract in pub api
@@ -212,13 +217,10 @@ It accepts connections and will naturally route hakuban communication, like ever
 	* rate limiting
 	* periodic rebalance
 	* drop multiple-tags-per-object support in favour of exactly-one-tag-per-object (groups, possibly tree-structured)
-	* change msgpack to something faster: [bench](https://github.com/only-cliches/NoProto#benchmarks) [bincode](https://crates.io/crates/bincode)
+	* events with payloads (at least assignment_id)
+	* expose state for current assignment_id could return assignment_id that was used
 
 * bindings
-	* js client for browser (wasm)
-		* find out how to structure the project
-		* implement basic object access
-	* js client for nodejs
 
 */
 
