@@ -1,9 +1,10 @@
 use std::{collections::{HashMap, VecDeque}, sync::{atomic::Ordering, Arc, Mutex}};
 
+use instant::Instant;
 use log::trace;
 
 use super::{core::NodeCore, local::LocalNodeShared};
-use crate::{diff, events::{Action, Event, EventEmitter, EventMap, EventMerge, EventMergeInput}, message::{Change, DataChange, Message, ObjectChange, OptionsChange, TagChange}, object::{core::{ObjectRawData, ObjectVersion}, remote::RemoteObject}, tag::remote::RemoteTag, Descriptor, LocalNode, ObjectDescriptor, ObjectType, TagDescriptor};
+use crate::{diff, events::{Action, Event, EventEmitter, EventMap, EventMerge, EventMergeInput}, message::{Change, DataChange, Message, ObjectChange, OptionsChange, Preference, TagChange}, object::{core::{ObjectRawData, ObjectVersion, Synchronized}, remote::RemoteObject}, tag::remote::RemoteTag, Descriptor, LocalNode, ObjectDescriptor, ObjectType, TagDescriptor};
 
 /*
 This file is a bit heavy with data structures, because there is a lot to keep track of.
@@ -26,7 +27,7 @@ pub(super) struct State {
 
 #[derive(Clone)]
 pub(super) struct OptionsState {
-	pub(super) diff_request: Option<bool>,
+	pub(super) diff_request: Preference,
 }
 
 pub(super) struct ObjectState {
@@ -36,8 +37,8 @@ pub(super) struct ObjectState {
 }
 
 pub(super) struct TagState {
-	pub(super) received: RemoteTagState,
-	pub(super) sent: RemoteTagState,
+	pub(super) received: ReceivedRemoteTagState,
+	pub(super) sent: SentRemoteTagState,
 	pub(super) local: LocalTagState,
 }
 
@@ -50,36 +51,42 @@ struct ArchivedSentDataVersion {
 	data: ObjectRawData,
 }
 
-#[derive(Debug)]
-struct ArchivedReceivedDataVersion {
-	data_version: ObjectVersion,
-	data_type: Option<ObjectType>, // None if data_version was only declared by sender, without sending data
-	data: Option<ObjectRawData>,   // None if remote side didn't send data and we don't know the data version submitted
-}
-
 
 pub(super) struct SentRemoteObjectState {
 	pub(super) observed: bool,
+	observed_true_sent_with_message_id: Option<u64>,
 	pub(super) exposed: bool,
 	data: VecDeque<ArchivedSentDataVersion>, // newest versions on the front
+	data_synchronized: Option<Synchronized>,
 	//TODO: this doesn't work
-	pub(super) cost: Option<Option<u32>>,
+	pub(super) cost: Option<u32>,
 }
 
 pub(super) struct ReceivedRemoteObjectState {
 	pub(super) observed: bool,
 	pub(super) exposed: bool,
-	data: VecDeque<ArchivedReceivedDataVersion>, // newest versions on the front
+	data: Option<ObjectRawData>,
+	data_version: Option<ObjectVersion>,
+	data_type: Option<ObjectType>,
+	data_last_synchronized_microseconds_ago: Option<Synchronized>,
 	//TODO: this doesn't work
-	pub(super) cost: Option<Option<u32>>,
+	pub(super) cost: Option<u32>,
 }
 
-pub(super) struct RemoteTagState {
+pub(super) struct SentRemoteTagState {
+	pub(super) observed: bool,
+	observed_true_sent_with_message_id: Option<u64>,
+	pub(super) exposed: bool,
+	pub(super) cost: Option<u32>,
+}
+
+pub(super) struct ReceivedRemoteTagState {
 	pub(super) observed: bool,
 	pub(super) exposed: bool,
-	pub(super) cost: Option<Option<u32>>,
+	pub(super) cost: Option<u32>,
 }
 
+
 pub(super) struct LocalObjectState {
 	pub(super) object_interface: RemoteObject,
 	pub(super) changes_token: Option<EventMergeInput<ObjectDescriptor>>,
@@ -105,8 +112,23 @@ impl State {
 	pub(super) fn object_borrow_mut(&mut self, descriptor: &ObjectDescriptor, local_node: &Arc<LocalNodeShared>, node: &Arc<NodeCore>) -> &mut ObjectState {
 		if !self.objects.contains_key(descriptor) {
 			let object_state = ObjectState {
-				received: ReceivedRemoteObjectState { observed: false, exposed: false, data: VecDeque::new(), cost: None },
-				sent: SentRemoteObjectState { observed: false, exposed: false, data: VecDeque::new(), cost: None },
+				received: ReceivedRemoteObjectState {
+					observed: false,
+					exposed: false,
+					data: None,
+					cost: None,
+					data_version: None,
+					data_type: None,
+					data_last_synchronized_microseconds_ago: None,
+				},
+				sent: SentRemoteObjectState {
+					observed: false,
+					observed_true_sent_with_message_id: None,
+					exposed: false,
+					data: VecDeque::new(),
+					data_synchronized: None,
+					cost: None,
+				},
 				local: LocalObjectState { object_interface: RemoteObject::new(local_node.clone(), node.clone(), descriptor), changes_token: None },
 			};
 			self.objects.insert(descriptor.clone(), object_state);
@@ -129,7 +151,6 @@ impl State {
 		// We need to keep those for event handler to know it needs to send observed=false and exposed=false.
 
 		let object_state = self.objects.get(descriptor).unwrap();
-		//dbg!(descriptor,object_state.sent.observed,object_state.sent.exposed,object_state.received.observed,object_state.received.exposed,remote_is_upstream,self.remote_is_tag_observer_of_object(descriptor), self.local_is_exposer_of_object(object_state),self.remote_is_tag_exposer_of_object(descriptor),self.local_is_observer_of_object(object_state));
 
 		if object_state.sent.observed
 			|| object_state.sent.exposed
@@ -157,7 +178,7 @@ impl State {
 		}
 	}
 
-	pub(super) fn tag_maybe_borrow(&mut self, descriptor: &TagDescriptor) -> Option<&mut TagState> {
+	pub(super) fn tag_maybe_borrow_mut(&mut self, descriptor: &TagDescriptor) -> Option<&mut TagState> {
 		if self.tags.contains_key(descriptor) {
 			Some(self.tags.get_mut(descriptor).unwrap())
 		} else {
@@ -168,8 +189,8 @@ impl State {
 	pub(super) fn tag_borrow_mut(&mut self, descriptor: &TagDescriptor, local_node: &Arc<LocalNodeShared>, node: &Arc<NodeCore>) -> &mut TagState {
 		if !self.tags.contains_key(descriptor) {
 			let tag_state = TagState {
-				received: RemoteTagState { observed: false, exposed: false, cost: None },
-				sent: RemoteTagState { observed: false, exposed: false, cost: None },
+				received: ReceivedRemoteTagState { observed: false, exposed: false, cost: None },
+				sent: SentRemoteTagState { observed: false, observed_true_sent_with_message_id: None, exposed: false, cost: None },
 				local: LocalTagState {
 					tag_interface: RemoteTag::new(local_node.clone(), node.clone(), descriptor),
 					changes_token: None,
@@ -248,22 +269,80 @@ impl State {
 	fn local_is_observer_of_tag(&self, tag_state: &TagState) -> bool {
 		tag_state.local.tag_interface.is_observed_by_other_node()
 	}
+
+	fn object_was_observed_at(&mut self, object_descriptor: &ObjectDescriptor, message_number: u64) -> bool {
+		if let Some(object) = self.object_maybe_borrow_mut(object_descriptor) {
+			if let Some(observed_true_sent_with_message_id) = object.sent.observed_true_sent_with_message_id {
+				if observed_true_sent_with_message_id <= message_number {
+					return true;
+				};
+			};
+		};
+		for tag_descriptor in &object_descriptor.tags {
+			if let Some(tag) = self.tag_maybe_borrow_mut(tag_descriptor) {
+				if let Some(observed_true_sent_with_message_id) = tag.sent.observed_true_sent_with_message_id {
+					if observed_true_sent_with_message_id <= message_number {
+						return true;
+					};
+				};
+			};
+		}
+		false
+	}
 }
 
 
 impl SentRemoteObjectState {
-	pub(super) fn apply_change(&mut self, change: &ObjectChange) {
+	pub(super) fn apply_change(&mut self, change: &ObjectChange, message_sequence_number: u64) {
+		//TODO: rename observed -> observer?
+		if !self.observed && change.observer.unwrap_or(false) {
+			self.observed_true_sent_with_message_id = Some(message_sequence_number);
+		}
 		self.observed = change.observer.unwrap_or(self.observed);
 		self.exposed = change.exposer.unwrap_or(self.exposed);
-		self.cost = change.cost.or(self.cost);
+		//TEST-MISSING: this "if" was missing and no test noticed it (detected while testing js binding manually. happend on creation on tag exposer on js side, with ruby observer waiting for object)
+		if self.observed || self.exposed {
+			self.cost = change.cost.or(self.cost);
+		} else {
+			self.cost = None;
+		}
 	}
 }
 
-impl RemoteTagState {
-	pub(super) fn apply_change(&mut self, change: &TagChange) {
+impl SentRemoteTagState {
+	pub(super) fn apply_change(&mut self, change: &TagChange, message_sequence_number: u64) {
+		//TEST-MISSING: this "if" was missing and no test noticed it (detected with manual object-expose + tag-observe test)
+		if !self.observed && change.observer.unwrap_or(false) {
+			self.observed_true_sent_with_message_id = Some(message_sequence_number);
+		}
 		self.observed = change.observer.unwrap_or(self.observed);
 		self.exposed = change.exposer.unwrap_or(self.exposed);
-		//TODO: cost? self.cost = change.cost.or(self.cost);
+		if self.observed || self.exposed {
+			self.cost = change.cost.or(self.cost);
+		} else {
+			self.cost = None;
+		}
+	}
+}
+
+impl ObjectState {
+	fn get_data(&self, requested_data_version: &Option<ObjectVersion>) -> Result<(Option<ObjectType>, ObjectRawData), String> {
+		match requested_data_version {
+			Some(requested_data_version) =>
+				if let Some(archived_data) = self.sent.data.iter().find(|archived_data| &archived_data.data_version == requested_data_version) {
+					Ok((Some(archived_data.data_type.clone()), archived_data.data.clone()))
+				} else if self.received.data.is_some() && self.received.data_version.as_ref() == Some(requested_data_version) {
+					//TODO: unwrap? really?
+					Ok((self.received.data_type.clone(), self.received.data.as_ref().unwrap().clone()))
+				} else {
+					//this can happen if remote side sent us data while we lost interest in the object
+					//TEST-NEEDED
+					//TODO: think if we should retain data for this case
+					Err("Can't locate data".to_string())
+				},
+
+			None => Ok((None, Arc::new(vec![]))),
+		}
 	}
 }
 
@@ -276,8 +355,8 @@ pub struct RemoteNodeMutable {
 	last_received_message_sequence_number: u64,
 	// last sent acknowledgment
 	last_sent_message_acknowledgement_number: u64,
-	diff_produce: Option<bool>,
-	diff_request: Option<bool>,
+	diff_produce: Preference,
+	diff_request: Preference,
 }
 
 pub struct RemoteNode {
@@ -322,7 +401,13 @@ impl RemoteNode {
 		local_node_events.include(object_events_map).forever();
 		local_node_events.include(tag_events_map).forever();
 
-		let state = State { objects: HashMap::new(), tags: HashMap::new(), object_changes, tag_changes, options: OptionsState { diff_request: None } };
+		let state = State {
+			objects: HashMap::new(),
+			tags: HashMap::new(),
+			object_changes,
+			tag_changes,
+			options: OptionsState { diff_request: Preference::NoPreference },
+		};
 		RemoteNode {
 			node_core: NodeCore::new(upstream, Some(0), Some(0), None),
 			local_node: local_node.shared.clone(),
@@ -332,8 +417,8 @@ impl RemoteNode {
 				last_sent_message_sequence_number: 0,
 				last_received_message_sequence_number: 0,
 				last_sent_message_acknowledgement_number: 0,
-				diff_produce,
-				diff_request,
+				diff_produce: diff_produce.into(),
+				diff_request: diff_request.into(),
 			}),
 		}
 	}
@@ -365,7 +450,7 @@ impl RemoteNode {
 		let mut mutable = self.mutable.lock().unwrap();
 		mutable.state.objects.clear();
 		mutable.state.tags.clear();
-		mutable.state.options.diff_request = None;
+		mutable.state.options.diff_request = Preference::NoPreference;
 		mutable.last_sent_message_sequence_number = 0;
 		mutable.last_received_message_sequence_number = 0;
 		mutable.last_sent_message_acknowledgement_number = 0;
@@ -414,10 +499,14 @@ impl RemoteNode {
 						mutable.state.options.diff_request = diff_request;
 					}
 					//TODO: custom data
+					//TODO: trigger event
 				}
 
 				Change::Object(object_change) => {
-					let object_state = mutable.state.object_borrow_mut(&object_change.descriptor, &self.local_node, &self.node_core);
+					let descriptor = ObjectDescriptor::from_json_strings(object_change.descriptor_tags, object_change.descriptor_json);
+					let was_observed_at = mutable.state.object_was_observed_at(&descriptor, message.acknowledgment_number); //TODO: move this ahead. it's only here because of borrow confilict.
+					let object_state = mutable.state.object_borrow_mut(&descriptor, &self.local_node, &self.node_core);
+
 					if let Some(observer) = object_change.observer {
 						if observer == object_state.received.observed {
 							return Err("Protocol error. Unneeded data received (object observer field).".to_string());
@@ -425,87 +514,75 @@ impl RemoteNode {
 						object_state.local.object_interface.observe(observer);
 						object_state.received.observed = observer;
 					}
+
 					if let Some(exposer) = object_change.exposer {
 						if exposer == object_state.received.exposed {
 							return Err("Protocol error. Unneeded data received (object exposer field).".to_string());
 						};
-						let cost = if exposer { object_change.cost.flatten().or_else(|| object_state.received.cost.flatten()).expect("No cost?") } else { 0 };
+						let cost = if exposer { object_change.cost.or(object_state.received.cost).expect("No cost?") } else { 0 };
 						object_state.local.object_interface.expose(exposer, cost);
 						object_state.received.exposed = exposer;
 					}
-					if let Some(data_version) = &object_change.data_version {
-						//TODO: fail if same or lower data_version received?
-						//TODO: fail if data_type received without data
-						if let Some(data) = object_change.data {
-							match data {
-								DataChange::Data { data } => {
-									let previous_archived_data = object_state.received.data.pop_back();
-									object_state.received.data.push_front(ArchivedReceivedDataVersion {
-										data_type: Some(object_change.data_type.or_else(|| previous_archived_data.unwrap().data_type).unwrap()),
-										data_version: data_version.clone(),
-										data: Some(data),
-									});
+
+					if object_change.data.is_some() || object_change.data_version.is_some() || object_change.data_last_synchronized_microseconds_ago.is_some() {
+						//TODO: declone
+						let latest_received_version = object_change.data_version.clone().or_else(|| object_state.received.data_version.clone()).unwrap();
+						let latest_received_data_type = object_change.data_type.clone().or_else(|| object_state.received.data_type.clone());
+
+						let synchronized = object_change
+							.data_last_synchronized_microseconds_ago
+							.map(|microseconds| match microseconds {
+								0 if was_observed_at => Synchronized::Now,
+								#[allow(unused_mut)] //for wasm
+								mut microseconds => {
+									// this sucks :(
+									// on wasm, Instant::now() is internally a low float, sometimes lower than the "microseconds", and we get:
+									// panicked at 'overflow when subtracting durations', library/core/src/time.rs:988:31
+									// so:
+									#[cfg(feature = "wasm")]
+									if microseconds > (instant::now() * 1000.0) as u64 {
+										microseconds = (instant::now() * 1000.0) as u64
+									}
+
+									Synchronized::LastAt(Instant::now() - std::time::Duration::from_micros(microseconds))
 								}
-								DataChange::Diff { diff, old_data_version } => {
-									let previous_archived_data = object_state.received.data.pop_back();
-									let empty = Arc::new(vec![]);
-									let (old_data, old_data_type) = if let Some(old_data_version) = old_data_version {
-										if let Some(archived_data) =
-											object_state.sent.data.iter().find(|archived_data| *archived_data.data_version == old_data_version)
-										{
-											(&archived_data.data, Some(&archived_data.data_type))
-										} else if previous_archived_data.is_some()
-											&& previous_archived_data.as_ref().unwrap().data.is_some()
-											&& *previous_archived_data.as_ref().unwrap().data_version == old_data_version
-										{
-											(
-												previous_archived_data.as_ref().unwrap().data.as_ref().unwrap(),
-												Some(previous_archived_data.as_ref().unwrap().data_type.as_ref().unwrap()),
-											)
-										} else {
-											//this can happen if remote side sent us data while we lost interest in the object
-											//TEST-NEEDED
-											//TODO: think if we should retain data for this case
-											continue;
-										}
-									} else {
-										(&empty, None)
-									};
-									object_state.received.data.push_front(ArchivedReceivedDataVersion {
-										data_version: data_version.clone(),
-										data: Some(Arc::new(diff::patch(old_data, &diff)?)),
-										data_type: Some(object_change.data_type.or_else(|| old_data_type.cloned()).unwrap()),
-									});
-								}
-							};
-							if object_state.local.object_interface.is_exposed_by_this_node() {
-								//TODO: or should we check if node is the **assigned** exposer?
+							})
+							.or_else(|| object_state.received.data_last_synchronized_microseconds_ago.clone());
+
+						let latest_received_data = match object_change.data {
+							Some(DataChange::Data { data }) => Some(data),
+							Some(DataChange::Diff { diff, old_data_version }) => {
+								let (_old_data_type, old_data) = object_state.get_data(&old_data_version)?;
+								Some(Arc::new(diff::patch(&old_data, &diff)?))
+							}
+							None => object_state.received.data.clone(),
+						};
+
+						if object_state.local.object_interface.is_exposed_by_this_node() {
+							if let Some(data) = &latest_received_data {
 								object_state.local.object_interface.state_set(
-									data_version,
-									object_state.received.data.front().as_ref().unwrap().data.as_ref().unwrap().clone(),
-									object_state.received.data.front().as_ref().unwrap().data_type.as_ref().unwrap().clone(),
+									&latest_received_version,
+									data.clone(),
+									latest_received_data_type.clone().unwrap(),
+									synchronized.clone().unwrap(),
 								);
 							} else {
-								return Err("Remote state error. Data submitted for object remote is not a exposer of.".to_string());
+								//this happens in first "expose" message if submitter already knows some version
 							}
-						} else {
-							//TODO: if there is no data sent, should that only be permitted for initial message? only for observers?
-							object_state.local.object_interface.maybe_synchronize(data_version);
-							let _previous_archived_data = object_state.received.data.pop_back();
-							object_state.received.data.push_front(ArchivedReceivedDataVersion {
-								data_version: data_version.clone(),
-								data_type: None,
-								data: None, //TODO: lookup locally
-							});
 						}
-					} else if object_change.data.is_some() || object_change.data_type.is_some() {
-						return Err("Remote state error. Data or data type submitted without data version.".to_string());
+
+						object_state.received.data = latest_received_data;
+						object_state.received.data_type = latest_received_data_type;
+						object_state.received.data_version = Some(latest_received_version);
+						object_state.received.data_last_synchronized_microseconds_ago = synchronized;
 					}
-					mutable.state.object_return(&object_change.descriptor, self.node_core.upstream);
+
+					mutable.state.object_return(&descriptor, self.node_core.upstream);
 				}
 
 				Change::Tag(tag_change) => {
-					let tag_state = mutable.state.tag_borrow_mut(&tag_change.descriptor, &self.local_node, &self.node_core);
+					let descriptor = TagDescriptor::from_json_string(tag_change.descriptor_json);
+					let tag_state = mutable.state.tag_borrow_mut(&descriptor, &self.local_node, &self.node_core);
 					if let Some(observer) = tag_change.observer {
 						if observer == tag_state.received.observed {
 							return Err("Protocol error. Unneeded data received (tag observer field).".to_string());
@@ -517,11 +594,11 @@ impl RemoteNode {
 						if exposer == tag_state.received.exposed {
 							return Err("Protocol error. Unneeded data received (tag exposer field).".to_string());
 						};
-						let cost = if exposer { tag_change.cost.flatten().or_else(|| tag_state.received.cost.flatten()).expect("No cost?") } else { 0 };
+						let cost = if exposer { tag_change.cost.or(tag_state.received.cost).expect("No cost?") } else { 0 };
 						tag_state.local.tag_interface.expose(exposer, cost);
 						tag_state.received.exposed = exposer;
 					};
-					mutable.state.tag_return(&tag_change.descriptor, self.node_core.upstream);
+					mutable.state.tag_return(&descriptor, self.node_core.upstream);
 				}
 			}
 		}
@@ -560,30 +637,20 @@ impl RemoteNode {
 		}
 	}
 
+	fn next_sent_message_sequence_number(&self, mutable: &RemoteNodeMutable) -> u64 {
+		mutable.last_sent_message_sequence_number + 1
+	}
+
 	fn process_object_event(&self, mutable: &mut RemoteNodeMutable, descriptor: ObjectDescriptor, action: Action) -> Result<Vec<Change>, String> {
 		trace!("{:?}, object event received {:?} {:?}", self.node_core, &action, &descriptor);
 		let mut changes = vec![];
 		match action {
 			Action::Insert | Action::Change => {
 				{
-					let options = mutable.state.options.clone();
+					let next_sent_message_sequence_number = self.next_sent_message_sequence_number(mutable);
+					let options = mutable.state.options.clone(); //meh
 					let object_state = mutable.state.object_borrow_mut(&descriptor, &self.local_node, &self.node_core);
 
-					let last_received_version = object_state.received.data.front();
-					let last_sent_version = object_state.sent.data.iter().find(|archived_version| archived_version.data_transmitted);
-					let highest_version_known_by_remote = vec![
-						last_received_version.as_ref().map(|archived| &archived.data_version),
-						last_sent_version.as_ref().map(|archived| &archived.data_version),
-					]
-					.into_iter()
-					.flatten()
-					.max();
-
-					let (_synchronized, data_version, data_type, data) = object_state.local.object_interface.state();
-					let local_node_knows_newer_version_than_remote = data_version
-						.as_ref()
-						.and_then(|newest_version| highest_version_known_by_remote.as_ref().map(|known_version| newest_version > known_version).or(Some(true)))
-						.unwrap_or(false);
 
 					let (declare_local_as_observer, declare_local_as_exposer) = if self.node_core.upstream {
 						(
@@ -598,86 +665,124 @@ impl RemoteNode {
 						)
 					};
 
-					let declare_cost = if declare_local_as_exposer { object_state.local.object_interface.minimum_cost_in_other_nodes() } else { None };
 
-					let (declare_data, data_type_of_declared_data) = if object_state.local.object_interface.is_assigned_to_other_node()
-						&& object_state.local.object_interface.is_observed_by_this_node()
-						&& local_node_knows_newer_version_than_remote
-					{
-						if mutable.diff_produce.unwrap_or_else(|| options.diff_request.unwrap_or(true)) {
-							let (newest_shared_data, newest_shared_data_type, newest_shared_data_version) = match (last_received_version, last_sent_version) {
-								(None, None) => (None, None, None),
-								(Some(received_version), None) =>
-									if received_version.data.is_some() {
-										(
-											Some(received_version.data.as_ref().unwrap()),
-											Some(received_version.data_type.as_ref().unwrap()),
-											Some(&received_version.data_version),
-										)
-									} else {
-										(None, None, None)
-									},
-								(None, Some(sent_version)) => (Some(&sent_version.data), Some(&sent_version.data_type), Some(&sent_version.data_version)),
-								(Some(received_version), Some(sent_version)) =>
-									if received_version.data_version > sent_version.data_version && received_version.data.is_some() {
-										(
-											Some(received_version.data.as_ref().unwrap()),
-											Some(received_version.data_type.as_ref().unwrap()),
-											Some(&received_version.data_version),
-										)
-									} else {
-										(Some(&sent_version.data), Some(&sent_version.data_type), Some(&sent_version.data_version))
-									},
-							};
-							let diff = object_state.local.object_interface.get_diff(
-								newest_shared_data_version,
-								newest_shared_data,
-								data_version.as_ref().unwrap(),
-								data.as_ref().unwrap(),
-							);
-							if let Some(diff) = diff {
-								(
-									Some(DataChange::Diff { diff, old_data_version: newest_shared_data_version.cloned() }),
-									newest_shared_data_type.or_else(|| Some(data_type.as_ref().unwrap())),
-								)
+					let remote_wants_data_and_type =
+						object_state.local.object_interface.is_observed_by_this_node() && !object_state.local.object_interface.is_assigned_to_this_node();
+
+					let declare_cost =
+						if declare_local_as_exposer { Some(object_state.local.object_interface.minimum_cost_in_other_nodes().unwrap()) } else { None };
+					let local_state = object_state.local.object_interface.state();
+					let last_sent_version = object_state.sent.data.iter().find(|archived_version| archived_version.data_transmitted);
+					let declare_data_version = local_state.as_ref().map(|state| state.version.clone());
+
+					//TODO: maybe use declare_local_as_exposer instead?
+					let (declare_data, declare_data_type) = if remote_wants_data_and_type && local_state.is_some() {
+						//TODO: less clones!
+						let last_received_version = object_state.received.data_version.clone();
+						let highest_version_known_by_remote = vec![
+							&last_received_version, //.as_ref().map(|archived| &archived.data_version),
+							&last_sent_version.as_ref().map(|archived| archived.data_version.clone()),
+						]
+						.into_iter()
+						.flatten()
+						.max()
+						.cloned();
+
+						let local_node_knows_newer_version_than_remote = match (&local_state, highest_version_known_by_remote) {
+							(None, None) => false,
+							(None, Some(_)) => false,
+							(Some(_), None) => true,
+							(Some(local_state), Some(remote_version)) => local_state.version > remote_version,
+						};
+
+						if local_node_knows_newer_version_than_remote {
+							let local_state = local_state.as_ref().unwrap();
+							if mutable.diff_produce.or(options.diff_request).unwrap_or(true) {
+								let (newest_shared_data, newest_shared_data_type, newest_shared_data_version) =
+									match (&last_received_version, last_sent_version) {
+										(None, None) => (None, None, None),
+										(Some(_received_version), None) =>
+											if object_state.received.data.is_some() {
+												(
+													object_state.received.data.as_ref(),
+													object_state.received.data_type.as_ref(),
+													object_state.received.data_version.as_ref(),
+												)
+											} else {
+												(None, None, None)
+											},
+										(None, Some(sent_version)) =>
+											(Some(&sent_version.data), Some(&sent_version.data_type), Some(&sent_version.data_version)),
+										(Some(received_version), Some(sent_version)) =>
+											if received_version > &sent_version.data_version && object_state.received.data.is_some() {
+												(
+													object_state.received.data.as_ref(),
+													object_state.received.data_type.as_ref(),
+													object_state.received.data_version.as_ref(),
+												)
+											} else {
+												(Some(&sent_version.data), Some(&sent_version.data_type), Some(&sent_version.data_version))
+											},
+									};
+								let diff = object_state.local.object_interface.get_diff(
+									newest_shared_data_version,
+									newest_shared_data,
+									&local_state.version,
+									&local_state.raw,
+								);
+								if let Some(diff) = diff {
+									(
+										Some(DataChange::Diff { diff, old_data_version: newest_shared_data_version.cloned() }),
+										newest_shared_data_type.or(Some(&local_state.data_type)),
+									)
+								} else {
+									(Some(DataChange::Data { data: local_state.raw.clone() }), Some(&local_state.data_type))
+								}
 							} else {
-								(Some(DataChange::Data { data: data.clone().unwrap() }), Some(data_type.as_ref().unwrap()))
+								(Some(DataChange::Data { data: local_state.raw.clone() }), Some(&local_state.data_type))
 							}
 						} else {
-							(Some(DataChange::Data { data: data.clone().unwrap() }), Some(data_type.as_ref().unwrap()))
+							(None, None)
 						}
 					} else {
 						(None, None)
 					};
 
-					let declare_data_type = data_type_of_declared_data
+					let declare_data_type = declare_data_type
 						.filter(|new_data_type| last_sent_version.map(|last_sent_version| last_sent_version.data_type != **new_data_type).unwrap_or(true));
 
+					let should_send_data_sync = (remote_wants_data_and_type || (declare_local_as_exposer && !object_state.sent.exposed))
+						&& object_state.sent.data_synchronized.as_ref() != local_state.as_ref().map(|local_state| &local_state.synchronized);
+
+					let should_send_data_version = (remote_wants_data_and_type
+						|| ((declare_local_as_observer && !object_state.sent.observed) || (declare_local_as_exposer && !object_state.sent.exposed)))
+						&& declare_data_version != last_sent_version.as_ref().map(|data| data.data_version.clone());
+
 					let change = ObjectChange {
-						descriptor: descriptor.clone(),
+						descriptor_tags: descriptor.tags_to_strings(),
+						descriptor_json: descriptor.json_to_string(),
 						observer: if declare_local_as_observer != object_state.sent.observed { Some(declare_local_as_observer) } else { None },
 						exposer: if declare_local_as_exposer != object_state.sent.exposed { Some(declare_local_as_exposer) } else { None },
-						data_version: if declare_data.is_some()
-							|| ((declare_local_as_observer && !object_state.sent.observed) || (declare_local_as_exposer && !object_state.sent.exposed))
-						{
-							data_version.clone()
+						cost: if declare_cost != object_state.sent.cost { declare_cost } else { None },
+						data_version: if should_send_data_version { declare_data_version } else { None },
+						data_type: declare_data_type.cloned(),
+						data_last_synchronized_microseconds_ago: if should_send_data_sync {
+							local_state.as_ref().map(|state| state.synchronized.micros_ago())
 						} else {
 							None
 						},
-						data_type: declare_data_type.cloned(),
 						data: declare_data,
-						cost: if Some(declare_cost) != object_state.sent.cost { Some(declare_cost) } else { None },
 					};
-					object_state.sent.apply_change(&change);
+					object_state.sent.apply_change(&change, next_sent_message_sequence_number);
 
 					if !change.is_empty() {
 						if change.data_version.is_some() {
 							object_state.sent.data.push_front(ArchivedSentDataVersion {
 								can_be_dropped_at_ack: mutable.last_sent_message_sequence_number,
-								data_version: data_version.unwrap(),
+								data_version: local_state.as_ref().unwrap().version.clone(),
 								data_transmitted: change.data.is_some(),
-								data_type: data_type.unwrap(),
-								data: data.unwrap(),
+								data_type: local_state.as_ref().unwrap().data_type.clone(),
+								data: local_state.unwrap().raw,
 							});
 						}
 						changes.push(Change::Object(change));
@@ -686,19 +791,22 @@ impl RemoteNode {
 				mutable.state.object_return(&descriptor, self.node_core.upstream);
 			}
 			Action::Remove => {
+				let next_sent_message_sequence_number = self.next_sent_message_sequence_number(mutable);
 				if let Some(object_state) = mutable.state.object_maybe_borrow_mut(&descriptor) {
 					let change = ObjectChange {
-						descriptor: descriptor.clone(),
+						descriptor_tags: descriptor.tags_to_strings(),
+						descriptor_json: descriptor.json_to_string(),
 						observer: if object_state.sent.observed { Some(false) } else { None },
 						exposer: if object_state.sent.exposed { Some(false) } else { None },
 						//TODO: should we clear all following fields?
 						data: None,
 						data_type: None,
 						data_version: None,
+						data_last_synchronized_microseconds_ago: None,
 						cost: None,
 					};
 
-					object_state.sent.apply_change(&change);
+					object_state.sent.apply_change(&change, next_sent_message_sequence_number);
 					mutable.state.object_return(&descriptor, self.node_core.upstream);
 
 					if !change.is_empty() {
@@ -715,6 +823,7 @@ impl RemoteNode {
 		if self.node_core.upstream {
 			match action {
 				Action::Insert | Action::Change => {
+					let next_sent_message_sequence_number = self.next_sent_message_sequence_number(mutable);
 					let tag_state = mutable.state.tag_borrow_mut(&descriptor, &self.local_node, &self.node_core);
 
 					let declare_local_as_observer = tag_state.local.tag_interface.is_observed_by_other_node();
@@ -722,36 +831,37 @@ impl RemoteNode {
 					let declare_cost = if declare_local_as_exposer { tag_state.local.tag_interface.minimum_cost_in_other_nodes() } else { None };
 
 					let change = TagChange {
-						descriptor: descriptor.clone(),
+						descriptor_json: descriptor.json_to_string(),
 						observer: if declare_local_as_observer != tag_state.sent.observed { Some(declare_local_as_observer) } else { None },
 						exposer: if declare_local_as_exposer != tag_state.sent.exposed { Some(declare_local_as_exposer) } else { None },
-						cost: if Some(declare_cost) != tag_state.sent.cost { Some(declare_cost) } else { None },
+						cost: if declare_cost != tag_state.sent.cost { declare_cost } else { None },
 					};
 
-					tag_state.sent.apply_change(&change);
+					tag_state.sent.apply_change(&change, next_sent_message_sequence_number);
 					mutable.state.tag_return(&descriptor, self.node_core.upstream);
 
 					if !change.is_empty() {
 						changes.push(Change::Tag(change))
 					};
 				}
-				Action::Remove =>
-					if let Some(tag_state) = mutable.state.tag_maybe_borrow(&descriptor) {
+				Action::Remove => {
+					let next_sent_message_sequence_number = self.next_sent_message_sequence_number(mutable);
+					if let Some(tag_state) = mutable.state.tag_maybe_borrow_mut(&descriptor) {
 						let change = TagChange {
-							descriptor: descriptor.clone(),
+							descriptor_json: descriptor.json_to_string(),
 							observer: if tag_state.sent.observed { Some(false) } else { None },
 							exposer: if tag_state.sent.exposed { Some(false) } else { None },
-							//TODO: should we clear this explicitly?
 							cost: None,
 						};
 
-						tag_state.sent.apply_change(&change);
+						tag_state.sent.apply_change(&change, next_sent_message_sequence_number);
 						mutable.state.tag_return(&descriptor, self.node_core.upstream);
 
 						if !change.is_empty() {
 							changes.push(Change::Tag(change));
 						};
-					},
+					}
+				}
 			}
 		}
 		Ok(changes)
@@ -764,12 +874,14 @@ impl RemoteNode {
 		for (descriptor, state) in &mutable.state.objects {
 			ret.push_str(
 				format!(
-					"│ {:?} Received({}) Sent({}) ReceivedVersionsStored:{}  SentVersionsStored:{}\n",
+					"│ {:?} Received({}) Sent({}) ReceivedVersionsStored:{}  SentVersionsStored:{}  ReceivedSentCost:{},{}\n",
 					&descriptor,
 					vec![if state.received.observed { "observed" } else { "" }, if state.received.exposed { "exposed" } else { "" }].join(","),
 					vec![if state.sent.observed { "observed" } else { "" }, if state.sent.exposed { "exposed" } else { "" }].join(","),
-					state.received.data.len(),
-					state.sent.data.len()
+					state.received.data.is_some(),
+					state.sent.data.len(),
+					state.received.cost.map(|value| value.to_string()).unwrap_or_else(|| "".to_string()),
+					state.sent.cost.map(|value| value.to_string()).unwrap_or_else(|| "".to_string()),
 				)
 				.as_str(),
 			);
