@@ -1,6 +1,8 @@
-use std::{collections::{HashMap, HashSet}, hash::{Hash, Hasher}, sync::{Arc, Mutex, RwLock, RwLockWriteGuard}, time::SystemTime};
+use std::{cmp::{max, Ordering}, collections::{HashMap, HashSet}, hash::{Hash, Hasher}, sync::{Arc, Mutex, RwLock, RwLockWriteGuard}, time::SystemTime};
 
-use crate::{descriptor::ObjectDescriptor, diff, events::EventSource, expose_contract::ExposeContract, node::{core::NodeCore, LocalNodeShared}, observe_contract::ObserveContract, tag::core::TagCore, TagDescriptor};
+use instant::Instant;
+
+use crate::{descriptor::ObjectDescriptor, diff, events::EventSource, expose_contract::ExposeContract, node::{core::NodeCore, LocalNodeShared}, observe_contract::ObserveContract, tag::core::TagCore};
 
 /// `= Vec<i64>`
 pub type ObjectVersion = Vec<i64>;
@@ -16,8 +18,8 @@ pub(crate) struct ObjectCore {
 	pub descriptor: ObjectDescriptor,
 	pub(super) diffs: Mutex<HashMap<VersionFromTo, Arc<Mutex<Option<ObjectRawData>>>>>,
 	pub(super) diff_states: Mutex<HashMap<ObjectVersion, Arc<Mutex<Option<diff::State>>>>>,
-	tags: HashMap<TagDescriptor, Arc<TagCore>>,
-	state: RwLock<ObjectState>,
+	tags: Vec<Arc<TagCore>>,
+	state: RwLock<Option<ObjectState>>,
 	links: RwLock<ObjectLinks>,
 	changes: EventSource<ObjectDescriptor>,
 }
@@ -29,24 +31,37 @@ pub(crate) struct ObjectLinks {
 	assigned_expose_contract: Option<Arc<dyn ExposeContract>>,
 }
 
+#[derive(Clone, Debug, PartialEq, Eq)]
+pub(crate) enum Synchronized {
+	//Never,
+	LastAt(Instant), // data was exposed by somebody but "Now" conditions were not met. TODO: what should happen if newest version gets uploaded by non-current-exposer?
+	Now,             // (some, not necessarily current?) data was exposed by currently assigned exposer while exposer was aware of the most recent assignment id
+}
+
+impl Synchronized {
+	pub fn micros_ago(&self) -> u64 {
+		match self {
+			Synchronized::LastAt(timestamp) => max(1, (Instant::now().duration_since(*timestamp)).as_micros() as u64),
+			Synchronized::Now => 0,
+		}
+	}
+}
+
+
+#[derive(Clone)]
 pub(crate) struct ObjectState {
-	data: Option<RawData>,
-	synchronized: bool,
+	pub raw: ObjectRawData,
+	pub data_type: ObjectType,
+	pub version: ObjectVersion,
+	pub synchronized: Synchronized,
 }
 
-struct RawData {
-	raw: ObjectRawData,
-	data_type: ObjectType,
-	version: ObjectVersion,
-}
-
-
 
 impl ObjectLinks {
 	fn recalculate_active(&mut self, object_core: &Arc<ObjectCore>) -> bool {
 		let mut has_tag_observe_contracts = false;
 		let mut has_tag_expose_contracts = false;
-		for tag in object_core.tags.values() {
+		for tag in object_core.tags.iter() {
 			let (tag_observe_contracts, tag_expose_contracts) = tag.has_observe_expose_contracts();
 			has_tag_observe_contracts = has_tag_observe_contracts || tag_observe_contracts;
 			has_tag_expose_contracts = has_tag_expose_contracts || tag_expose_contracts;
@@ -60,7 +75,7 @@ impl ObjectLinks {
 			|| !self.expose_contracts.is_empty() && (!self.observe_contracts.is_empty() || has_tag_observe_contracts);
 
 		if self.active != previous_active {
-			for tag in object_core.tags.values() {
+			for tag in object_core.tags.iter() {
 				if self.active {
 					tag.active_insert(object_core.clone());
 				} else {
@@ -72,7 +87,7 @@ impl ObjectLinks {
 		if self.active && self.assigned_expose_contract.is_none() {
 			self.expose_contract_assign(object_core);
 		} else if !self.active && self.assigned_expose_contract.is_some() {
-			self.expose_contract_unassign(object_core);
+			self.current_expose_contract_unassign(object_core);
 		};
 
 		//this is always triggered, not only when if active has changed, because despite object not changing, the tags may have changed
@@ -100,7 +115,7 @@ impl ObjectLinks {
 				}
 			}
 
-			for tag in object_core.tags.values() {
+			for tag in object_core.tags.iter() {
 				for candidate_expose_contract in tag.expose_contracts() {
 					if let Some(load) = candidate_expose_contract.expected_proportional_load_with() {
 						let candidate_score = (!candidate_expose_contract.node_core().upstream, load);
@@ -128,7 +143,7 @@ impl ObjectLinks {
 			let current_score = (!current_expose_contract.node_core().upstream, current_expose_contract.expected_proportional_load_without());
 			let candidate_score = (!candidate_contract.node_core().upstream, candidate_contract.expected_proportional_load_without());
 			if current_score > candidate_score && candidate_contract.assign(object_core.clone()) {
-				current_expose_contract.unassign(&object_core.descriptor);
+				Self::expose_contract_unassign(object_core, current_expose_contract);
 				self.assigned_expose_contract = Some(candidate_contract);
 			}
 		} else if candidate_contract.assign(object_core.clone()) {
@@ -136,26 +151,23 @@ impl ObjectLinks {
 		}
 	}
 
-	fn expose_contract_unassign(&mut self, object_core: &ObjectCore) -> bool {
-		match &self.assigned_expose_contract {
+	fn expose_contract_unassign(object_core: &ObjectCore, contract_to_unassign: &Arc<dyn ExposeContract>) {
+		contract_to_unassign.unassign(&object_core.descriptor);
+		if let Some(object_state) = &mut *object_core.state.write().unwrap() {
+			object_state.synchronized = Synchronized::LastAt(Instant::now());
+		};
+	}
+
+	fn current_expose_contract_unassign(&mut self, object_core: &ObjectCore) -> bool {
+		match self.assigned_expose_contract.as_ref() {
 			Some(assigned_expose_contract) => {
-				assigned_expose_contract.unassign(&object_core.descriptor);
+				Self::expose_contract_unassign(object_core, assigned_expose_contract);
 				self.assigned_expose_contract = None;
-				object_core.state.write().unwrap().synchronized = false;
 				true
 			}
 			None => false,
 		}
 	}
-
-	//TODO: this already exists somewhere else
-	fn is_assigned_to_node(&self, node: &Arc<NodeCore>) -> bool {
-		if let Some(ref expose_contract) = self.assigned_expose_contract {
-			expose_contract.node_core() == node
-		} else {
-			false
-		}
-	}
 }
 
 impl Hash for ObjectCore {
@@ -174,11 +186,11 @@ impl Eq for ObjectCore {}
 
 
 impl ObjectCore {
-	pub(crate) fn new(local_node: Arc<LocalNodeShared>, descriptor: ObjectDescriptor, tags: HashMap<TagDescriptor, Arc<TagCore>>) -> ObjectCore {
+	pub(crate) fn new(local_node: Arc<LocalNodeShared>, descriptor: ObjectDescriptor, tags: Vec<Arc<TagCore>>) -> ObjectCore {
 		ObjectCore {
 			local_node,
 			descriptor: descriptor.clone(),
-			state: RwLock::new(ObjectState { data: None, synchronized: false }),
+			state: RwLock::new(None),
 			links: RwLock::new(ObjectLinks {
 				observe_contracts: HashMap::new(),
 				expose_contracts: HashMap::new(),
@@ -196,20 +208,18 @@ impl ObjectCore {
 		let links = links.unwrap_or_else(|| self.links.write().unwrap());
 		let state = self.state.read().unwrap();
 		format!(
-			"├─────\n│ Object: {:?}\n│ Active: {:?}\n│ Observe contracts: {:?}\n│ Expose contracts: {:?}\n│ Tags: {:?}\n│ Assigned: {:?}\n│ Synchronized: {:?}\n| Diffs stored: {}  diff states stored: {}  Arc refs: {:?}\n",
+			"├─────\n│ Object: {:?}\n│ Active: {:?}\n│ Observe contracts: {:?}\n│ Expose contracts: {:?}\n│ Tags: {:?}\n│ Assigned: {:?}\n| Diffs stored: {}  diff states stored: {}\n",
 			self.descriptor,
 			links.active,
 			links.observe_contracts.keys(),
 			links.expose_contracts.keys(),
 			self.tags,
-			links.assigned_expose_contract,
-			state.synchronized,
+			links.assigned_expose_contract.as_ref().map(|contract| contract.node_core()),
 			self.diffs.lock().unwrap().len(),
 			self.diff_states.lock().unwrap().len(),
-			self.state.read().unwrap().data.as_ref().map(|raw_data| Arc::strong_count(&raw_data.raw))
-		) + &match &state.data {
+		) + &match &*state {
 			Some(data) => {
-				format!("│ Data: {:?} {:?} {:?}B\n", data.version, data.data_type, data.raw.len())
+				format!("│ Data: {:?} {:?} {:?}B {:?}\n", data.version, data.data_type, data.raw.len(), data.synchronized)
 			}
 			None => "│ Data: None\n".to_string(),
 		}
@@ -243,7 +253,7 @@ impl ObjectCore {
 		};
 		if let Some(ref expose_contract) = links.assigned_expose_contract {
 			if expose_contract.id() == contract.id() {
-				links.expose_contract_unassign(self);
+				links.current_expose_contract_unassign(self);
 			}
 		};
 		links.recalculate_active(self);
@@ -265,73 +275,127 @@ impl ObjectCore {
 		let mut links = self.links.write().unwrap();
 		if let Some(ref expose_contract) = links.assigned_expose_contract {
 			if expose_contract.id() == contract.id() {
-				links.expose_contract_unassign(self);
+				links.current_expose_contract_unassign(self);
 			}
 		}
 	}
 
-	pub(crate) fn state_get(&self) -> (bool, Option<ObjectVersion>, Option<ObjectType>, Option<ObjectRawData>) {
-		let locked_state = &self.state.read().unwrap();
-		if let Some(ref data) = locked_state.data {
-			(locked_state.synchronized, Some(data.version.clone()), Some(data.data_type.clone()), Some(data.raw.clone()))
-		} else {
-			(locked_state.synchronized, None, None, None)
-		}
+	pub(crate) fn state_get(&self) -> Option<ObjectState> {
+		let locked_state = self.state.read().unwrap();
+		(*locked_state).clone()
 	}
 
 	pub(crate) fn version_get(&self) -> Option<ObjectVersion> {
 		let locked_state = &self.state.read().unwrap();
-		locked_state.data.as_ref().map(|data| data.version.clone())
+		locked_state.as_ref().map(|data| data.version.clone())
 	}
 
 	#[allow(clippy::ptr_arg)]
-	pub(crate) fn state_set(&self, node: &Arc<NodeCore>, version: &ObjectVersion, data_type: ObjectType, raw: ObjectRawData) -> bool {
-		let links = self.links.read().unwrap();
-		let mut locked_state = self.state.write().unwrap();
-		let fresh = if let Some(current_version) = &locked_state.data { current_version.version < *version } else { true }; //TODO: prettify when let can be used with other conditions
-		if fresh {
-			locked_state.data = Some(RawData { raw, data_type, version: version.clone() });
-			self.synchronize_if_assigned(&links, &mut locked_state, node);
+	pub fn state_set_with_precalculated_synchronized(
+		&self, mut locked_state: RwLockWriteGuard<Option<ObjectState>>, version: &ObjectVersion, data_type: ObjectType, raw: ObjectRawData,
+		new_synchronized: Synchronized,
+	) -> bool {
+		let submitted_version_is = locked_state.as_ref().map(|state| version.cmp(&state.version)).unwrap_or(Ordering::Greater);
+		if submitted_version_is == Ordering::Greater {
+			*locked_state = Some(ObjectState { raw, data_type, version: version.clone(), synchronized: new_synchronized });
 			self.changes.change();
-		}
-		fresh
-	}
+			return true;
+		};
 
-	pub(crate) fn version_set(self: &Arc<Self>, node: &Arc<NodeCore>, mut version: Option<ObjectVersion>) -> bool {
-		let links = self.links.read().unwrap();
-		let mut locked_state = self.state.write().unwrap();
-		//TODO: prettify when let can be used with other conditions
-		let fresh = if let Some(new_version) = &version {
-			if let Some(current_version) = &locked_state.data {
-				current_version.version < *new_version
-			} else {
-				true
+		if let Some(state) = &mut *locked_state {
+			if state.synchronized != new_synchronized {
+				state.synchronized = new_synchronized;
+				self.changes.change();
 			}
 		} else {
-			let previous_version = locked_state.data.as_ref().map(|data| data.version.clone());
-			version = Some(Self::version_next(previous_version));
-			true
-		};
-		if fresh {
-			if let Some(raw_data) = &mut locked_state.data {
-				raw_data.version = version.unwrap();
-			} else {
-				panic!(); //TODO: maybe not? but what if not? error? return false?
-			}
-			self.synchronize_if_assigned(&links, &mut locked_state, node);
-			self.changes.change();
+			panic!();
 		}
-		fresh
+
+		false
 	}
 
-	pub(crate) fn data_set(self: &Arc<Self>, node: &Arc<NodeCore>, data_type: ObjectType, raw: ObjectRawData) {
+	#[allow(clippy::ptr_arg)]
+	pub(crate) fn state_set_with_node(
+		&self, node: &Arc<NodeCore>, version: &ObjectVersion, data_type: ObjectType, raw: ObjectRawData, synchronized: Synchronized,
+	) -> bool {
+		let links = self.links.read().unwrap();
+		let locked_state = self.state.write().unwrap();
+		let new_synchronized = self.new_synchronized_state(
+			Some(node.id()) == links.assigned_expose_contract.as_ref().map(|assigned| assigned.node_core().id()),
+			&locked_state,
+			synchronized,
+		);
+		self.state_set_with_precalculated_synchronized(locked_state, version, data_type, raw, new_synchronized)
+	}
+
+	#[allow(clippy::ptr_arg)]
+	pub(crate) fn state_set_with_assignment_id(
+		&self, version: &ObjectVersion, data_type: ObjectType, raw: ObjectRawData, assignment_id: u64, synchronized: Synchronized,
+	) -> bool {
+		let links = self.links.read().unwrap();
+		let locked_state = self.state.write().unwrap();
+		let new_synchronized = self.new_synchronized_state(
+			Some(Some(assignment_id)) == links.assigned_expose_contract.as_ref().map(|assigned| assigned.assignment(&self.descriptor)),
+			&locked_state,
+			synchronized,
+		);
+		self.state_set_with_precalculated_synchronized(locked_state, version, data_type, raw, new_synchronized)
+	}
+
+	fn new_synchronized_state(
+		&self, assigned: bool, locked_state: &RwLockWriteGuard<Option<ObjectState>>, proposed_synchronized: Synchronized,
+	) -> Synchronized {
+		//TODO: check if sync timestamp is not in the future
+		match locked_state.as_ref().map(|state| state.synchronized.clone()) {
+			None => match proposed_synchronized {
+				Synchronized::LastAt(_new_timestamp) => proposed_synchronized,
+				Synchronized::Now =>
+					if assigned {
+						Synchronized::Now
+					} else {
+						Synchronized::LastAt(Instant::now())
+					},
+			},
+			Some(Synchronized::LastAt(old_timestamp)) => match proposed_synchronized {
+				Synchronized::LastAt(new_timestamp) => Synchronized::LastAt(max(new_timestamp, old_timestamp)),
+				Synchronized::Now =>
+					if assigned {
+						Synchronized::Now
+					} else {
+						Synchronized::LastAt(max(Instant::now(), old_timestamp)) //hmmm...
+					},
+			},
+			Some(Synchronized::Now) =>
+				if assigned {
+					proposed_synchronized //TEST-MISSING: this branch did not exist, always leaving synchronized as Now. router failed to accept and propagate desynchronize coming from wasm downstream.
+				} else {
+					Synchronized::Now
+				},
+		}
+	}
+
+	pub(crate) fn desynchronize(&self, assignment_id: u64) {
 		let links = self.links.read().unwrap();
 		let mut locked_state = self.state.write().unwrap();
-		let previous_version = locked_state.data.as_ref().map(|data| data.version.clone());
-		let version = Self::version_next(previous_version);
-		locked_state.data = Some(RawData { raw, data_type, version });
-		self.synchronize_if_assigned(&links, &mut locked_state, node);
-		self.changes.change();
+		if let Some(ref mut state) = *locked_state {
+			if Some(Some(assignment_id)) == links.assigned_expose_contract.as_ref().map(|assigned| assigned.assignment(&self.descriptor)) {
+				state.synchronized = Synchronized::LastAt(Instant::now());
+				self.changes.change();
+			}
+		}
+	}
+
+	//TODO: should we check contract too, in addition to node?
+	pub(crate) fn data_set(self: &Arc<Self>, node: &Arc<NodeCore>, data_type: ObjectType, raw: ObjectRawData) {
+		let links = self.links.read().unwrap();
+		let locked_state = self.state.write().unwrap();
+		let version = Self::version_next(locked_state.as_ref().map(|state| state.version.clone()));
+		let new_synchronized = self.new_synchronized_state(
+			Some(node.id()) == links.assigned_expose_contract.as_ref().map(|assigned| assigned.node_core().id()),
+			&locked_state,
+			Synchronized::Now,
+		);
+		self.state_set_with_precalculated_synchronized(locked_state, &version, data_type, raw, new_synchronized);
 	}
 
 	//TODO: maybe create a PROPER Version type, and put this in it?
@@ -352,27 +416,8 @@ impl ObjectCore {
 		}
 	}
 
-	pub(super) fn synchronize_if_assigned(&self, links: &ObjectLinks, state: &mut ObjectState, node: &Arc<NodeCore>) -> bool {
-		let previous_synchronized = state.synchronized;
-		if links.is_assigned_to_node(node) {
-			state.synchronized = true;
-		}
-		state.synchronized != previous_synchronized
-	}
-
-	// only used in object/remote.rs
-	pub(super) fn synchronize_if_assigned_and_version_matches(self: &Arc<Self>, node: &Arc<NodeCore>, version: &[i64]) {
-		let links = self.links.read().unwrap();
-		let mut locked_state = self.state.write().unwrap();
-		if let Some(current_version) = &locked_state.data {
-			if current_version.version == *version && self.synchronize_if_assigned(&links, &mut locked_state, node) {
-				self.changes.change();
-			}
-		}
-	}
-
-	pub(crate) fn synchronized(&self) -> bool {
-		self.state.read().unwrap().synchronized
+	pub(crate) fn synchronized(&self) -> Option<Synchronized> {
+		self.state.read().unwrap().as_ref().map(|state| state.synchronized.clone())
 	}
 
 	pub(crate) fn tag_changed(self: &Arc<Self>) {
@@ -387,14 +432,14 @@ impl ObjectCore {
 		let object_links = self.links.read().unwrap();
 		object_links.observe_contracts.len() > 1
 			|| (!object_links.observe_contracts.is_empty() && !object_links.observe_contracts.contains_key(node))
-			|| (check_tags && self.tags.values().any(|tag| tag.has_observer_other_than(node)))
+			|| (check_tags && self.tags.iter().any(|tag| tag.has_observer_other_than(node)))
 	}
 
 	pub(super) fn has_exposer_other_than(self: &Arc<Self>, node: &Arc<NodeCore>, check_tags: bool) -> bool {
 		let object_links = self.links.read().unwrap();
 		object_links.expose_contracts.len() > 1
 			|| (!object_links.expose_contracts.is_empty() && !object_links.expose_contracts.contains_key(node))
-			|| (check_tags && self.tags.values().any(|tag| tag.has_exposer_other_than(node)))
+			|| (check_tags && self.tags.iter().any(|tag| tag.has_exposer_other_than(node)))
 	}
 
 	pub(super) fn is_assigned_to_node(self: &Arc<Self>, node: &Arc<NodeCore>) -> bool {
@@ -417,43 +462,39 @@ impl ObjectCore {
 
 	pub(super) fn is_exposed_by_node(self: &Arc<Self>, node: &Arc<NodeCore>) -> bool {
 		let object_links = self.links.read().unwrap();
-		object_links.expose_contracts.contains_key(node) || self.tags.values().any(|tag| tag.is_exposed_by_node(node))
+		object_links.expose_contracts.contains_key(node) || self.tags.iter().any(|tag| tag.is_exposed_by_node(node))
 	}
 
 	pub(super) fn is_observed_by_node(self: &Arc<Self>, node: &Arc<NodeCore>) -> bool {
 		let object_links = self.links.read().unwrap();
-		object_links.observe_contracts.contains_key(node) || self.tags.values().any(|tag| tag.is_observed_by_node(node))
+		object_links.observe_contracts.contains_key(node) || self.tags.iter().any(|tag| tag.is_observed_by_node(node))
 	}
 
 	pub(super) fn minimum_cost_excluding_node(self: &Arc<Self>, node: &Arc<NodeCore>) -> Option<u32> {
 		let object_links = self.links.read().unwrap();
 
-		let mut lowest_cost = u32::MAX;
+		let mut lowest_cost = None;
 
 		for expose_contracts in object_links.expose_contracts.values() {
 			for expose_contract in expose_contracts {
 				if expose_contract.node_core() != node {
 					let cost = expose_contract.composition_cost();
-					if cost < lowest_cost {
-						lowest_cost = cost;
+					if lowest_cost.is_none() || cost < lowest_cost.unwrap() {
+						lowest_cost = Some(cost);
 					}
 				}
 			}
 		}
 
-		for tag in self.tags.values() {
+		for tag in self.tags.iter() {
 			if let Some(cost) = tag.minimum_cost_excluding_node(node) {
-				if cost < lowest_cost {
-					lowest_cost = cost;
+				if lowest_cost.is_none() || cost < lowest_cost.unwrap() {
+					lowest_cost = Some(cost);
 				}
 			}
 		}
 
-		if lowest_cost == u32::MAX {
-			None
-		} else {
-			Some(lowest_cost)
-		}
+		lowest_cost
 	}
 }
 
