@@ -1,8 +1,8 @@
 #![allow(clippy::mutex_atomic)] //because I want the value AND the locking. I think :)
 
-use std::{collections::hash_map::Entry, hash::{Hash, Hasher}, sync::{atomic::{AtomicBool, AtomicU32, Ordering}, Arc, Mutex}};
+use std::{collections::hash_map::Entry, hash::{Hash, Hasher}, sync::{atomic::{AtomicU32, Ordering}, Arc, Mutex, RwLock}};
 
-use super::core::{ObjectCore, ObjectRawData, ObjectType, ObjectVersion};
+use super::{core::{ObjectCore, ObjectRawData, ObjectState, ObjectType, ObjectVersion, Synchronized}, Assignment};
 use crate::{contract::Contract, diff, events::EventSource, expose_contract::ExposeContract, node::{core::NodeCore, local::LocalNodeShared}, observe_contract::ObserveContract, ObjectDescriptor};
 
 pub(crate) struct RemoteObject {
@@ -44,7 +44,7 @@ impl RemoteObject {
 		}
 	}
 
-	pub(crate) fn state(&self) -> (bool, Option<ObjectVersion>, Option<ObjectType>, Option<ObjectRawData>) {
+	pub(crate) fn state(&self) -> Option<ObjectState> {
 		self.object_interface.object_core.state_get()
 	}
 
@@ -76,13 +76,9 @@ impl RemoteObject {
 		self.object_interface.object_core.has_exposer_other_than(&self.object_interface.remote_node_core, check_tags)
 	}
 
-	pub(crate) fn maybe_synchronize(&self, version: &[i64]) {
-		self.object_interface.object_core.synchronize_if_assigned_and_version_matches(&self.object_interface.remote_node_core, version);
-	}
-
 	#[allow(clippy::ptr_arg)]
-	pub(crate) fn state_set(&self, version: &ObjectVersion, raw: ObjectRawData, data_type: ObjectType) -> bool {
-		self.object_interface.object_core.state_set(&self.object_interface.remote_node_core, version, data_type, raw)
+	pub(crate) fn state_set(&self, version: &ObjectVersion, raw: ObjectRawData, data_type: ObjectType, synchronized: Synchronized) -> bool {
+		self.object_interface.object_core.state_set_with_node(&self.object_interface.remote_node_core, version, data_type, raw, synchronized)
 	}
 
 	pub(crate) fn minimum_cost_in_other_nodes(&self) -> Option<u32> {
@@ -153,14 +149,13 @@ impl Drop for RemoteObject {
 }
 
 
-
 struct RemoteObjectInterface {
 	remote_node_core: Arc<NodeCore>,
 	object_core: Arc<ObjectCore>,
 	expose: Mutex<bool>,
 	observe: Mutex<bool>,
 	composition_cost: AtomicU32,
-	assigned: AtomicBool,
+	assignment: RwLock<Assignment>,
 	changes: EventSource<ObjectDescriptor>,
 }
 
@@ -173,7 +168,7 @@ impl RemoteObjectInterface {
 			expose: Mutex::new(false),
 			observe: Mutex::new(false),
 			composition_cost: AtomicU32::new(0),
-			assigned: AtomicBool::new(false),
+			assignment: RwLock::new(Assignment { counter: 0, assigned: false }),
 		})
 	}
 }
@@ -233,25 +228,37 @@ impl ExposeContract for RemoteObjectInterface {
 		self.composition_cost.load(Ordering::SeqCst) //TODO: weaken
 	}
 
-	fn assign(&self, _object: Arc<ObjectCore>) -> bool {
-		if self.assigned.swap(true, Ordering::Relaxed) {
+	fn assign(&self, _object_core: Arc<ObjectCore>) -> bool {
+		let mut assignment = self.assignment.write().unwrap();
+		if assignment.assigned {
 			panic!("Double assign of ObjectExpose");
 		};
-		self.node_core().load.fetch_add(self.composition_cost(), Ordering::SeqCst);
+		assignment.counter += 1;
+		assignment.assigned = true;
+		self.node_core().load.fetch_add(self.composition_cost(), Ordering::SeqCst); //TODO: maybe weaker?
 		self.changes.change();
-		true
+		true //TODO: do we need to check if we still exist?
 	}
 
 	fn unassign(&self, _object_descriptor: &ObjectDescriptor) -> bool {
-		if !self.assigned.swap(false, Ordering::Relaxed) {
+		//TODO: panic if object doesn't match?
+		let mut assignment = self.assignment.write().unwrap();
+		if !assignment.assigned {
 			panic!("Double unassign of ObjectExpose");
-		};
-		self.node_core().load.fetch_sub(self.composition_cost(), Ordering::SeqCst);
-		self.changes.change(); //FORMER-BUG, REGRESSION-TEST-NEEDED, this was missing and no test detected it. is it even really a problem?
-		true
+		}
+		assignment.assigned = false;
+		self.node_core().load.fetch_sub(self.composition_cost(), Ordering::SeqCst); //TODO: maybe weaker?
+		self.changes.change(); //FORMER-BUG, REGRESSION-TEST-NEEDED, this was missing and no rust test detected it
+		true //TODO: do we need to check if we still exist?
 	}
 
-	fn assigned(&self, _object_descriptor: &ObjectDescriptor) -> bool {
-		self.assigned.load(Ordering::Relaxed)
+	fn assignment(&self, _object_descriptor: &ObjectDescriptor) -> Option<u64> {
+		//TODO: panic if object doesn't match?
+		let assignment = self.assignment.read().unwrap();
+		if assignment.assigned {
+			Some(assignment.counter)
+		} else {
+			None
+		}
 	}
 }
