@@ -1,10 +1,11 @@
 /*! Provides Websocket transport running in Tokio runtime
 */
 
-use std::{cmp::min, collections::HashMap, convert::TryInto, sync::{Arc, Mutex, RwLock}, time::{Duration, Instant}};
+use std::{cmp::min, collections::HashMap, convert::TryInto, sync::{Arc, Mutex, RwLock}, time::Duration};
 
 use async_tungstenite::{tokio::connect_async, tungstenite};
 use futures::{stream::{SplitSink, SplitStream}, SinkExt, StreamExt};
+use instant::Instant;
 use log::{error, info, trace};
 use tokio::{net::TcpListener, time::{sleep, timeout}};
 use url::Url;
@@ -160,7 +161,6 @@ impl<E: std::fmt::Debug> std::error::Error for WebsocketListenerError<E> {}
 
 
 impl WebsocketListener {
-	#[must_use = "Newly created WebsocketListener needs to be held on to to keep listening"]
 	pub async fn new<T: TryInto<Url> + std::fmt::Debug>(local_node: LocalNode, address: T) -> Result<WebsocketListener, WebsocketListenerError<T::Error>>
 	where
 		T::Error: std::fmt::Debug,
@@ -559,7 +559,7 @@ where
 		let sink = sink_mutex.as_mut().unwrap();
 		if !messages.is_empty() {
 			for message in messages {
-				let tungstenite_message = tungstenite::Message::Binary(rmp_serde::to_vec(&message).unwrap());
+				let tungstenite_message = tungstenite::Message::Binary(message.serialize().unwrap());
 				sink.send(tungstenite_message).await.map_err(Termination::from_tungstenite_error)?;
 			}
 			*self.last_sent_at.lock().unwrap() = Instant::now(); //FORMER-BUG, REGRESSION-TEST-NEEDED: this was called even if messages vector was empty
@@ -596,10 +596,10 @@ where
 			// trying to receive something
 			if let Ok(message) = timeout(min(self.parameters.timeout / 4, Duration::from_millis(1000)), ws_stream_read.as_mut().unwrap().next()).await {
 				match message {
-					Some(Ok(tungstenite::Message::Binary(message_msgpack))) => {
+					Some(Ok(tungstenite::Message::Binary(serialized_message))) => {
 						*self.last_received_at.lock().unwrap() = Instant::now();
-						if !message_msgpack.is_empty() {
-							return rmp_serde::from_read_ref::<Vec<u8>, Message>(&message_msgpack).map_err(Termination::from_other_error);
+						if !serialized_message.is_empty() {
+							return Message::deserialize(&serialized_message).map_err(Termination::from_other_error);
 						}
 					}
 					Some(Ok(tungstenite::Message::Ping(_payload))) => {
