@@ -41,22 +41,21 @@ impl MockClient {
 							TestAction::NetJoin => {
 								ws_stream = Some(async_tungstenite::tokio::connect_async(target_address.as_ref().unwrap().clone()).await.unwrap().0);
 							}
-							TestAction::Send(message) => {
-								//println!("sending message {:?}", message);
-								match ws_stream {
-									Some(ref mut stream) => stream.send(tungstenite::Message::Binary(rmp_serde::to_vec(&message).unwrap())).await.unwrap(),
-									None => panic!(),
-								};
-							}
+							TestAction::Send(messages) =>
+								for message in messages {
+									match ws_stream {
+										Some(ref mut stream) => stream.send(tungstenite::Message::Binary(bincode::serialize(&message).unwrap())).await.unwrap(),
+										None => panic!(),
+									};
+								},
 							TestAction::Receive(mut expected_messages) => {
-								//println!("waiting for message {:?}", expected_message);
 								match ws_stream {
-									Some(ref mut stream) => {
+									Some(ref mut stream) =>
 										while !expected_messages.is_empty() {
 											tokio::select! {
 												next_message = stream.next() => match next_message {
 													Some(Ok(tungstenite::Message::Binary(message_msgpack))) => {
-														let received_changes = rmp_serde::from_read_ref::<Vec<u8>, Message>(&message_msgpack).unwrap().changes;
+														let received_changes = bincode::deserialize::<Message>(&message_msgpack).unwrap().changes;
 														let found_at = expected_messages
 															.iter()
 															.enumerate()
@@ -69,7 +68,6 @@ impl MockClient {
 																		return false;
 																	};
 																}
-																//to_send.push();
 																true
 															})
 															.unwrap_or_else(|| panic!("❌ERROR: {} received message different than expected.\nExpected:\n{:?}\nReceived:\n{:?}\n", name, expected_messages, received_changes))
@@ -83,8 +81,7 @@ impl MockClient {
 													panic!("\n❌{} - TIMEOUT waiting for messages {:?}", name, expected_messages);
 												}
 											};
-										}
-									}
+										},
 									None => panic!(),
 								};
 							}
@@ -95,7 +92,7 @@ impl MockClient {
 											Some(Ok(tungstenite::Message::Ping(_))) => { },
 											Some(Ok(tungstenite::Message::Close(_))) => { break; },
 											Some(Ok(tungstenite::Message::Binary(message_msgpack))) => {
-												let received_message = rmp_serde::from_read_ref::<Vec<u8>,Message>(&message_msgpack).unwrap();
+												let received_message = bincode::deserialize::<Message>(&message_msgpack).unwrap();
 												panic!("\n❌ERROR: {} received message while expecting disconnect. message: {:?}", name, received_message);
 											},
 											other => panic!("\n❌ERROR: {} received unexpected event while expecting disconnect. event: {:?}", name, other),
@@ -109,19 +106,24 @@ impl MockClient {
 							},
 							TestAction::ReceiveNothing(duration) => {
 								match ws_stream {
-									Some(ref mut stream) => tokio::select! {
-										_ = tokio::time::sleep(Duration::from_secs_f32(duration)) => {},
-										message = stream.next() => match message {
-											Some(Ok(tungstenite::Message::Binary(message_msgpack))) => {
-												let received_message = rmp_serde::from_read_ref::<Vec<u8>,Message>(&message_msgpack).unwrap();
-												panic!("\n❌ERROR: {} received message, while it should not. Message: {:?}",&name, &received_message);
-											},
-											_ => panic!()
+									Some(ref mut stream) => loop {
+										tokio::select! {
+											_ = tokio::time::sleep(Duration::from_secs_f32(duration)) => break,
+											message = stream.next() => match message {
+												Some(Ok(tungstenite::Message::Binary(message_msgpack))) => {
+													let received_message = bincode::deserialize::<Message>(&message_msgpack).unwrap();
+													if !received_message.changes.is_empty() {
+														panic!("\n❌ERROR: {} received message, while it should not. Message: {:?}",&name, &received_message);
+													};
+												},
+												_ => panic!()
+											}
 										}
 									},
 									None => panic!(),
 								};
 							}
+
 							TestAction::Sleep(duration) => {
 								tokio::time::sleep(Duration::from_secs_f32(duration)).await;
 							}
