@@ -1,4 +1,4 @@
-use std::{collections::HashMap, sync::{atomic::Ordering, Arc}, time::Duration};
+use std::{collections::HashMap, sync::{atomic::Ordering, Arc}, time::{Duration, Instant}};
 
 extern crate hakuban;
 
@@ -58,7 +58,9 @@ impl RealClient {
 								drop(connector.take());
 								match client_stream.as_mut().unwrap().next().await.unwrap().unwrap() {
 									tungstenite::Message::Close(_) => {}
-									message => panic!("Got message instea of Close: {:?}", message),
+									tungstenite::Message::Binary(message_msgpack) =>
+										panic!("Got message instead of Close: {:?}", bincode::deserialize::<Message>(&message_msgpack).unwrap()),
+									message => panic!("Got message instead of Close: {:?}", message),
 								}
 								target_stream.take().unwrap().close(None).await.unwrap();
 							}
@@ -71,32 +73,58 @@ impl RealClient {
 								client_stream = Some(async_tungstenite::tokio::accept_async(tcp_client_stream).await.unwrap());
 								target_stream = Some(async_tungstenite::tokio::connect_async(target_address.as_ref().unwrap().clone()).await.unwrap().0);
 							}
-							TestAction::Send(expected_message) => match client_stream.as_mut().unwrap().next().await.unwrap() {
-								Ok(tungstenite::Message::Binary(message_msgpack)) => {
-									let client_changes = rmp_serde::from_read_ref::<Vec<u8>, Message>(&message_msgpack).unwrap().changes;
-									assert_eq!(expected_message.changes.len(), client_changes.len());
-									for (expected_change, client_change) in expected_message.changes.iter().zip(client_changes.iter()) {
-										assert_eq!(
-											expected_change, client_change,
-											"\n❌ERROR: {} sent change different than expected. Left is expected, Right is received:",
-											name
-										);
+							TestAction::Send(mut expected_messages) => {
+								match client_stream {
+									Some(ref mut stream) => {
+										let mut to_send = vec![];
+										while !expected_messages.is_empty() {
+											tokio::select! {
+												next_message = stream.next() => match next_message {
+													Some(Ok(tungstenite::Message::Binary(message_msgpack))) => {
+														let received_changes = bincode::deserialize::<Message>(&message_msgpack).unwrap().changes;
+														let found_at = expected_messages
+															.iter()
+															.enumerate()
+															.find(|(_, expected_message)| {
+																if expected_message.changes.len() != received_changes.len() {
+																	return false
+																}
+																for (expected_change, target_change) in expected_message.changes.iter().zip(received_changes.iter()) {
+																	if expected_change != target_change {
+																		return false;
+																	};
+																}
+																true
+															})
+															.unwrap_or_else(|| panic!("❌ERROR: {} sent message different than expected.\nExpected:\n{:?}\nSent:\n{:?}\n", name, expected_messages, received_changes))
+															.0;
+														expected_messages.remove(found_at);
+														to_send.push(message_msgpack);
+													},
+													Some(Err(error)) => panic!("Connection error: {:?}", error),
+													_ => panic!()
+												},
+												_ = tokio::time::sleep(Duration::from_secs_f32(1.0)) => {
+													panic!("\n❌{} - TIMEOUT waiting for messages {:?}", name, expected_messages);
+												}
+											};
+										}
+										for message_msgpack in to_send {
+											let tungstenite_message = tungstenite::Message::Binary(message_msgpack);
+											target_stream.as_mut().unwrap().send(tungstenite_message).await.unwrap();
+										}
 									}
-									let tungstenite_message = tungstenite::Message::Binary(
-										rmp_serde::to_vec(&Message { changes: client_changes, sequence_number, acknowledgment_number }).unwrap(),
-									);
-									target_stream.as_mut().unwrap().send(tungstenite_message).await.unwrap();
-								}
-								Ok(message) => panic!("Unexpected message: {:?}", message),
-								Err(error) => panic!("{:?}", error),
-							},
+									None => panic!(),
+								};
+							}
+
 							TestAction::SendSomeMessages { count } =>
 								for _ in 0..count {
 									match client_stream.as_mut().unwrap().next().await.unwrap() {
 										Ok(tungstenite::Message::Binary(message_msgpack)) => {
-											let client_changes = rmp_serde::from_read_ref::<Vec<u8>, Message>(&message_msgpack).unwrap().changes;
+											let client_changes = bincode::deserialize::<Message>(&message_msgpack).unwrap().changes;
 											let tungstenite_message = tungstenite::Message::Binary(
-												rmp_serde::to_vec(&Message { changes: client_changes, sequence_number, acknowledgment_number }).unwrap(),
+												bincode::serialize(&Message { changes: client_changes, sequence_number, acknowledgment_number }).unwrap(),
 											);
 											target_stream.as_mut().unwrap().send(tungstenite_message).await.unwrap();
 										}
@@ -112,7 +140,7 @@ impl RealClient {
 											tokio::select! {
 												next_message = stream.next() => match next_message {
 													Some(Ok(tungstenite::Message::Binary(message_msgpack))) => {
-														let received_changes = rmp_serde::from_read_ref::<Vec<u8>, Message>(&message_msgpack).unwrap().changes;
+														let received_changes = bincode::deserialize::<Message>(&message_msgpack).unwrap().changes;
 														let found_at = expected_messages
 															.iter()
 															.enumerate()
@@ -183,21 +211,27 @@ impl RealClient {
 									Err(_) => panic!("\n❌ERROR: {} got timeout while waiting to get disconnected", &name),
 								};
 							},
-							TestAction::ReceiveNothing(duration) =>
+							TestAction::ReceiveNothing(duration) => loop {
 								match timeout(Duration::from_secs_f32(duration), target_stream.as_mut().unwrap().next()).await {
-									Ok(Some(Ok(tungstenite::Message::Binary(received)))) => panic!(
-										"\n❌ERROR: {} received message, while it should not. Message: {:?}",
-										&name,
-										rmp_serde::from_read_ref::<Vec<u8>, Message>(&received).unwrap().changes
-									),
+									Ok(Some(Ok(tungstenite::Message::Binary(received)))) => {
+										let deserialized = bincode::deserialize::<Message>(&received).unwrap();
+										if !deserialized.changes.is_empty() {
+											panic!("\n❌ERROR: {} received message, while it should not. Message: {:?}", &name, deserialized);
+										};
+									}
 									Ok(Some(Ok(received))) => panic!("\n❌ERROR: {} received message, while it should not. Message: {:?}", &name, &received),
 									Ok(Some(Err(error))) => panic!("\n❌ERROR: {} receive errored out, while it should not. Error: {:?}", &name, &error),
 									Ok(None) => panic!(),
-									Err(_) => {}
-								},
+									Err(_) => break,
+								}
+							},
 							TestAction::SendNothing(duration) => match timeout(Duration::from_secs_f32(duration), client_stream.as_mut().unwrap().next()).await
 							{
-								Ok(Some(Ok(sent))) => panic!("\n❌ERROR: {} sent message, while it should not. Message: {:?}", &name, &sent),
+								Ok(Some(Ok(sent))) => panic!(
+									"\n❌ERROR: {} sent message, while it should not. Message: {:?}",
+									&name,
+									bincode::deserialize::<Message>(&sent.into_data()).unwrap()
+								),
 								Ok(Some(Err(error))) => panic!("\n❌ERROR: {} send errored out, while it should not. Error: {:?}", &name, &error),
 								Ok(None) => panic!(),
 								Err(_) => {}
@@ -240,7 +274,7 @@ impl RealClient {
 								let object = observed_objects.get(&descriptor).expect("Unknown descriptor. Broken test?");
 								assert_eq!(object.object_data().unwrap().unwrap(), data);
 							}
-							TestAction::ObjectObserveEvent { action, descriptor, version, data, data_type, synchronized } => {
+							TestAction::ObjectObserveEvent { action, descriptor, version, data, data_type, synchronized, last_synchronized_ago } => {
 								let _enter = span.enter();
 								let received_event =
 									observed_objects_event_buffers.get_mut(&descriptor).expect("Descriptor unknown. Test error?").next().await.unwrap();
@@ -249,20 +283,36 @@ impl RealClient {
 								let received_key = received_event.key;
 								if let Some(version) = version {
 									let object = observed_objects.get(&received_key).unwrap();
-									assert_eq!(version, object.object_state().unwrap().data_version.unwrap());
+									assert_eq!(version, object.object_state().unwrap().unwrap().data_version);
 								};
 								if let Some(data_type) = data_type {
 									let object = observed_objects.get(&received_key).unwrap();
-									assert_eq!(data_type, object.object_state().unwrap().data_type.unwrap());
+									assert_eq!(data_type, object.object_state().unwrap().unwrap().data_type);
 								};
 								if let Some(data) = data {
 									let object = observed_objects.get(&received_key).unwrap();
-									assert_eq!(data, object.object_state().unwrap().data.unwrap());
+									assert_eq!(data, object.object_state().unwrap().unwrap().data);
 								};
 								if let Some(synchronized) = synchronized {
 									let object = observed_objects.get(&received_key).unwrap();
-									assert_eq!(synchronized, object.synchronized());
+									if let Some(object_synchronized) = object.synchronized() {
+										if synchronized == 0 {
+											assert_eq!(object_synchronized, 0);
+										} else if object_synchronized > synchronized + 10000 || object_synchronized < synchronized - 10000 {
+											panic!();
+										}
+									} else {
+										panic!();
+									}
 								};
+								if let Some(last_synchronized) = last_synchronized_ago {
+									let object = observed_objects.get(&received_key).unwrap();
+									if last_synchronized == 0 {
+										assert_eq!(0, object.object_state().unwrap().unwrap().synchronized);
+									} else {
+										assert!((object.object_state().unwrap().unwrap().synchronized as i64 - last_synchronized as i64).abs() < 10000);
+									}
+								}
 							}
 							TestAction::TagObjectObserveEvent(events) => {
 								let _enter = span.enter();
@@ -290,7 +340,7 @@ impl RealClient {
 								assert_eq!(events.len(), matched_events_count);
 								//TODO: verify stuff
 							}
-							TestAction::ObjectExposeEvent { action, descriptor, assigned } => {
+							TestAction::ObjectExposeEvent { action, descriptor, assignment } => {
 								let _enter = span.enter();
 								let received_event = exposed_objects_event_buffers.get_mut(&descriptor).unwrap().next().await.unwrap();
 								let received_key = received_event.key.clone();
@@ -298,9 +348,9 @@ impl RealClient {
 								if received_key != descriptor {
 									panic!("wrong object changed. expected {:?}, changed {:?}", received_key, descriptor);
 								};
-								if let Some(assigned) = assigned {
+								if let Some(assignment) = assignment {
 									let object = exposed_objects.get(&received_key).unwrap();
-									assert_eq!(assigned, object.assigned());
+									assert_eq!(Some(assignment), object.assignment());
 								};
 							}
 							TestAction::TagObjectExposeEvent(events) => {
@@ -329,22 +379,37 @@ impl RealClient {
 								assert_eq!(events.len(), matched_events_count);
 							}
 							TestAction::ObjectVersionExposed { .. } => {}
-							TestAction::ObjectExposeNewVersion { descriptor, data_version, data, data_type } => {
+							TestAction::ObjectExposeNewVersion { descriptor, data_version, data, data_type, last_synchronized_ago } => {
 								let _enter = span.enter();
 								let object = exposed_objects.get(&descriptor).expect("Unknown descriptor. Broken test?");
 								if data_type != ["JSON"] {
 									panic!("we don't support any other serialization than JSON rn");
 								};
 								if let Some(data_version) = data_version {
-									object.set_object_state(&data_version, &data).unwrap();
+									if let Some(last_synchronized) = last_synchronized_ago {
+										if last_synchronized > 0 {
+											object
+												.set_object_state_at(
+													&data_version,
+													&data,
+													object.assignment().unwrap(),
+													Instant::now() - Duration::from_micros(last_synchronized),
+												)
+												.unwrap();
+										} else {
+											object.set_object_state(&data_version, &data, object.assignment().unwrap()).unwrap();
+										}
+									} else {
+										object.set_object_state(&data_version, &data, object.assignment().unwrap()).unwrap();
+									}
 								} else {
 									object.set_object_data(&data).unwrap();
 								}
 							}
-							TestAction::ObjectExposeAssigned { descriptor, assigned } => {
+							TestAction::ObjectExposeAssigned { descriptor, assignment_id } => {
 								let _enter = span.enter();
 								let object = exposed_objects.get(&descriptor).expect("Unknown descriptor. Broken test?");
-								assert_eq!(assigned, object.assigned());
+								assert_eq!(assignment_id, object.assignment());
 							}
 							TestAction::TagObjectExposeNewVersion { tag, descriptor, data_version, data, data_type } => {
 								let _enter = span.enter();
@@ -352,7 +417,7 @@ impl RealClient {
 								if data_type != ["JSON"] {
 									panic!("we don't support any other serialization than JSON rn");
 								};
-								tag.set_object_state(&descriptor, &data_version, &data).unwrap();
+								tag.set_object_state(&descriptor, &data_version, &data, tag.object_assignment(&descriptor).unwrap()).unwrap();
 							}
 							TestAction::TagObserveNew { descriptor } => {
 								let _enter = span.enter();
@@ -399,16 +464,6 @@ impl RealClient {
 								let tag = exposed_tags.get(&tag_descriptor).expect("Unknown descriptor. Broken test?");
 								tag.set_object_data(&object_descriptor, &data).unwrap();
 							}
-							TestAction::ObjectExposeVersion { descriptor, data_version } => {
-								let _enter = span.enter();
-								let object = exposed_objects.get(&descriptor).expect("Unknown descriptor. Broken test?");
-								assert!(object.set_object_version(data_version));
-							}
-							TestAction::TagExposeVersion { tag_descriptor, object_descriptor, data_version } => {
-								let _enter = span.enter();
-								let tag = exposed_tags.get(&tag_descriptor).expect("Unknown descriptor. Broken test?");
-								assert!(tag.set_object_version(&object_descriptor, data_version));
-							}
 							TestAction::ObjectObserveVersion { descriptor, data_version } => {
 								let _enter = span.enter();
 								let object = observed_objects.get(&descriptor).expect("Unknown descriptor. Broken test?");
